from typing import Dict, List, Optional, Tuple

from aiko_services.elements.media import image_io

from highlighter.agent.capabilities import Capability
from highlighter.agent.capabilities.base_capability import StreamEvent
from highlighter.core.data_models import DataSample

__all__ = ["ImageOverlay", "ImageResize"]


class ImageResize(Capability):
    class StreamParameters(Capability.StreamParameters):
        width: int = 0  # If zero, will maintain aspect ratio with respect to height
        height: int = 0  # If zero, will maintain aspect ratio with respect to width

    def process_frame(self, stream, data_samples: List[DataSample]) -> Tuple[StreamEvent, Optional[Dict]]:
        parameters = self.stream_parameters(stream.stream_id)
        data_samples = [ds.resize(width=parameters.width, height=parameters.height) for ds in data_samples]
        return StreamEvent.OKAY, {"data_samples": data_samples}


class ImageOverlay(image_io.ImageOverlay, Capability):

    def process_frame(self, stream, data_samples, annotations):
        for ds, annos in zip(data_samples, annotations):
            ds_anns = [a for a in annos if a.data_file_id == ds.data_file_id]
            overlay = ds.draw_annotations(ds_anns)
            ds.content = overlay
        return StreamEvent.OKAY, {"data_samples": data_samples}



"""Add recorded_until + legacy metadata columns to datafile

Revision ID: a1b2c3d4e5f6
Revises: 839d46550586
Create Date: 2025-11-28 08:15:00.000000

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "a1b2c3d4e5f6"
down_revision: Union[str, None] = "839d46550586"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Add recorded_until, legacy_image_id and mime_type columns."""

    op.add_column("datafile", sa.Column("recorded_until", sa.DateTime(), nullable=True))
    op.add_column("datafile", sa.Column("legacy_image_id", sa.Integer(), nullable=True))
    op.add_column("datafile", sa.Column("mime_type", sa.String(), nullable=True))


def downgrade() -> None:
    """Remove recorded_until, legacy_image_id and mime_type columns."""

    op.drop_column("datafile", "mime_type")
    op.drop_column("datafile", "legacy_image_id")
    op.drop_column("datafile", "recorded_until")



# Declaration order is based on the dependency on static references
#
# To Do
# ~~~~~
# - None, yet !

from .configuration import (
    create_password,
    get_hostname, get_mqtt_configuration, get_mqtt_host, get_mqtt_port,
    get_namespace, get_namespace_prefix, get_pid, get_username
)

from .context import ContextManager, get_context

from .graph import Graph, Node

from .importer import load_module, load_modules

from .lock import Lock

from .logger import (
    DEBUG, get_log_level_name, get_logger, LoggingHandlerMQTT,
    log_level_all, log_level_real, print_error
)

from .lru_cache import LRUCache

from .network import get_network_port_free, get_network_ports_used

from .parser import generate, parse, parse_float, parse_int, parse_number

from .system import (
    dir_base_name,
    get_memory_used, get_virtual_memory,
    print_memory_used, print_virtual_memory,
    get_uptime
)

from .utc_iso8601 import (
   datetime_epoch, datetime_now_utc_iso, epoch_to_utc_iso,
   local_iso_now, utc_iso_since_epoch, utc_iso_to_datetime,
   utc_iso_to_local
)



from collections.abc import Iterable
from datetime import datetime, timezone
from typing import Any, Dict, List, Tuple
from uuid import UUID

from highlighter import Entity, Observation, TaxonomyMappingRule
from highlighter.agent.capabilities.base_capability import Capability, StreamEvent
from highlighter.client.base_models.validated_set import ValidatedSet
from highlighter.client.gql_client import HLClient
from highlighter.core.labeled_uuid import LabeledUUID


class TaxonomyMapping(Capability):
    """
    Capability for applying taxonomy-mapping rules defined in Highlighter

    Example element definition:
    {
        "name": "Taxonomy Mapping",
        "input": [{ "name": "entities", "type": "Dict[UUID, Entity]" }],
        "output": [{ "name": "entities", "type": "Dict[UUID, Entity]" }],
        "deploy": { "local": { "module": "highlighter.agent.capabilities.taxonomy_mapping",
                               "class_name": "TaxonomyMapping" } },
        "parameters": {
            "mapping_rule_ids": [
                "2891fd3c-47a7-45a8-938f-7ed42f266a7c",
                "72f53bc7-1f24-4462-87de-9943c6fef931"
            },
        }
    }

    """

    class InitParameters(Capability.InitParameters):
        mapping_rule_ids: List[UUID]

    def __init__(self, context):
        super().__init__(context)
        all_rules = HLClient.get_client().entityAttributeMappingRules(return_type=List[TaxonomyMappingRule])
        self.rules = []
        for rule in all_rules:
            if rule.id in self.init_parameters.mapping_rule_ids:
                validate_rule(rule)
                self.rules.append(rule)

    def process_frame(self, stream, entities: Dict[UUID, Entity]) -> Tuple[StreamEvent, Dict]:
        if len(set(x.to_entity_attribute.id for x in self.rules)) < len(self.rules):
            raise ValueError("rules shouldn't have overlapping toEntityAttribute")

        for entity in entities.values():
            for rule in self.rules:
                for annotation in entity.annotations:
                    apply_rule(rule, annotation.observations)
                apply_rule(rule, entity.global_observations)

        # 'entities' will be changed by our modifications to the underlying observations
        return StreamEvent.OKAY, {"entities": entities}


def get_value(value, attributeEnum):
    if value is None or value == "":
        if attributeEnum is None:
            return None
        return attributeEnum.id
    return value


def validate_rule(rule: TaxonomyMappingRule):

    if rule.strategy == "direct":
        for map in rule.maps:
            if len(map.predicates) != 1:
                raise ValueError("map for direct strategy should only contain one predicate")
        from_entity_attributes = set(m.predicates[0].from_entity_attribute.id for m in rule.maps)
        if len(from_entity_attributes) != 1:
            raise ValueError(
                f"direct rule should only have one from_entity_attribute, {len(from_entity_attributes)} is received"
            )

        total_number_of_predicates = sum(1 for m in rule.maps for _ in m.predicates)
        unique_values_from_entity_attribute_values = set(
            (p.from_entity_attribute.id, get_value(p.from_value, p.from_entity_attribute_enum))
            for m in rule.maps
            for p in m.predicates
        )

        if len(unique_values_from_entity_attribute_values) < total_number_of_predicates:
            raise ValueError("duplicate values detected in direct rules")

    if rule.strategy == "most_confident":
        for m in rule.maps:
            from_entity_attributes = set(p.from_entity_attribute.id for p in m.predicates)
            if len(from_entity_attributes) != 1:
                raise ValueError(
                    f"each map should have only one from_entity_attribute, {len(from_entity_attributes)}"
                )

    # a rule at least have one predicate
    if len(rule.maps) == 0:
        raise ValueError("this rule has no maps")


def apply_rule(rule: TaxonomyMappingRule, observations: ValidatedSet[Observation]):
    """Apply taxonomy mapping rule"""

    from_attribute_ids = []
    rule.maps = sorted(rule.maps, key=lambda m: m.sort_order)
    for m in rule.maps:
        for p in m.predicates:
            if p.from_entity_attribute.id not in from_attribute_ids:
                from_attribute_ids.append(p.from_entity_attribute.id)

    if rule.strategy == "direct":
        d = {}
        for m in rule.maps:
            for p in m.predicates:
                d[(p.from_entity_attribute.id, get_value(p.from_value, p.from_entity_attribute_enum))] = (
                    get_value(m.to_value, m.to_entity_attribute_enum)
                )
                break

        for observation in observations:
            value = observation.value.id if hasattr(observation.value, "id") else observation.value
            result = d.get((observation.attribute_id, value))
            if result is not None:
                observation.attribute_id = rule.to_entity_attribute.id
                observation.value = result
                break

    elif rule.strategy == "rank":
        d = {}
        for m in rule.maps:
            for p in m.predicates:
                d[(p.from_entity_attribute.id, get_value(p.from_value, p.from_entity_attribute_enum))] = (
                    get_value(m.to_value, m.to_entity_attribute_enum)
                )
                break

        found = False
        for attribute_id in from_attribute_ids:
            if found:
                break
            for observation in observations:
                if observation.attribute_id == attribute_id:
                    value = observation.value
                    to_value = d.get((observation.attribute_id, value))
                    if to_value is not None:
                        observation.attribute_id = rule.to_entity_attribute.id
                        observation.value = to_value
                        found = True
                        break

    elif rule.strategy == "matched":
        d = {}
        for m in rule.maps:
            values = [None] * len(from_attribute_ids)
            for p in m.predicates:
                idx = from_attribute_ids.index(p.from_entity_attribute.id)
                values[idx] = get_value(p.from_value, p.from_entity_attribute_enum)
            d[tuple(values)] = get_value(m.to_value, m.to_entity_attribute_enum)

        _, value, _ = get_value_and_confidence_from_observations(observations, from_attribute_ids)

        for key, _value in d.items():
            found = True
            for left, right in zip(key, value):
                if (left != right) and left is not None:
                    found = False
                    break
            if found:
                observations.add(
                    Observation(
                        attribute_id=rule.to_entity_attribute.id,
                        value=_value,
                        datum_source=observations[0].datum_source,
                        occurred_at=datetime.now(timezone.utc),
                    )
                )
                break

    elif rule.strategy == "most_confident":
        d = {}
        for m in rule.maps:
            for p in m.predicates:
                d[(p.from_entity_attribute.id, get_value(p.from_value, p.from_entity_attribute_enum))] = (
                    get_value(m.to_value, m.to_entity_attribute_enum)
                )
                break

        obs, value, confidence = get_value_and_confidence_from_observations(observations, from_attribute_ids)
        idxes_by_confidence = sorted(list(range(len(confidence))), key=lambda x: -confidence[x])
        sorted_values = [value[i] for i in idxes_by_confidence]
        sorted_attrs = [from_attribute_ids[i] for i in idxes_by_confidence]
        sorted_obs = [obs[i] for i in idxes_by_confidence]
        for idx, attr, value, obs in zip(idxes_by_confidence, sorted_attrs, sorted_values, sorted_obs):
            result = d.get((attr, value))
            if result is not None:
                obs.attribute_id = rule.to_entity_attribute.id
                obs.value = result
                break

    else:
        raise ValueError(f"Unknown rule strategy {rule.strategy} and no default value provided")

    if (rule.default_value not in [None, ""]) or (rule.default_entity_attribute_enum is not None):
        observations.add(
            Observation(
                attribute_id=rule.to_entity_attribute.id,
                value=get_value(rule.default_value, rule.default_entity_attribute_enum),
                datum_source=observations[0].datum_source,
                occurred_at=datetime.now(timezone.utc),
            )
        )

    # Remove unmapped observations
    for observation in observations:
        if observation.attribute_id in from_attribute_ids:
            observations.remove(observation)


def get_value_and_confidence_from_observations(
    observations: Iterable[Observation], attr_ids: List[UUID]
) -> Tuple[List[Observation], List[Any], List[float]]:
    obs, value, confidence = [None] * len(attr_ids), [None] * len(attr_ids), [0.0] * len(attr_ids)
    for idx, attr_id in enumerate(attr_ids):
        for observation in observations:
            if observation.attribute_id == attr_id:
                value[idx] = observation.value
                value[idx] = getattr(value[idx], "id", value[idx])
                confidence[idx] = observation.datum_source.confidence
                obs[idx] = observation
                break
    return obs, value, confidence



from datetime import datetime
from typing import List, Optional
from uuid import UUID

from ...core import GQLBaseModel
from .page_info import PageInfo

__all__ = ["File", "FilePresigned", "FileConnection"]


class File(GQLBaseModel):
    """GraphQL File type response model"""

    id: str
    uuid: UUID
    content_type: str
    width: Optional[int] = None
    height: Optional[int] = None
    original_source_url: str
    mime_type: Optional[str] = None
    recorded_at: Optional[datetime] = None


class FilePresigned(File):
    file_url_original: str


class FileConnection(GQLBaseModel):
    page_info: PageInfo
    nodes: List[FilePresigned]



import re

from cookiecutter.utils import simple_filter

__all__ = ["camel_to_snake"]


@simple_filter
def camel_to_snake(name):
    name = str(name)
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()



# Usage
# ~~~~~
# $ python
# >>> import utc_iso8601
# >>> utc_iso8601.datetime_epoch()[0]
#     datetime.datetime(1970, 1, 1, 0, 0)
# >>> utc_iso8601.datetime_epoch()[1]
#     '1970-01-01T00:00:00.000000'
# >>> utc_iso8601.datetime_now_utc_iso()
#     '2024-10-29T11:31:22.000000'
# >>> utc_iso8601.epoch_to_utc_iso(1.0)
#     '1970-01-01T00:00:01'
# >>> utc_iso8601.local_iso_now()
#     '2024-10-29 11:31:22'
# >>> utc_iso8601.utc_iso_since_epoch("1970-01-01T00:00:01")
#     1.0
# >>> utc_iso8601.utc_iso_to_datetime("1970-01-01T00:00:01")
#     datetime.datetime(1970, 1, 1, 0, 0, 1)
# >>> utc_iso8601.utc_iso_to_local("1970-01-01T00:00:01")
#     '1970-01-01 10:00:01'
#
# Reference
# ~~~~~~~~~
# Note 1: Epoch is 1970-01-01 00:00:00
# Note 2: "datetime format" is that used by the Python "datatime" module
# Note 3: UTC ISO 8601 combined date and time: "yyyy-mm-ddThh:mm:ss.ssssss"
#
# https://en.wikipedia.org/wiki/ISO_8601#Combined_date_and_time_representations
#
# datetime_epoch()[0]: Epoch in datetime format
# datetime_epoch()[1]: Epoch in UTC ISO format
#
# datetime_now_utc_iso(): Current date/time in UTC ISO format
#
# epoch_to_utc_iso(seconds_since_epoch): Seconds since epoch in UTC ISO format
#
# local_iso_now(): Current date/time in local datetime format
#
# utc_iso_since_epoch(datetime_utc_iso): UTC ISO format in seconds since epoch
#
# utc_iso_to_datetime(datetime_utc_iso): UTC ISO format in date/time format
#
# utc_iso_to_local(datetime_utc_iso): UTC ISO format in local UTC ISO format
# 
# Notes
# ~~~~~
# No longer using utcnow and utcfromtimestamp
# - https://blog.ganssle.io/articles/2019/11/utcnow.html

from datetime import date, datetime, timezone

__all__ = [
   "datetime_epoch", "datetime_now_utc_iso", "epoch_to_utc_iso",
   "local_iso_now", "utc_iso_since_epoch", "utc_iso_to_datetime",
   "utc_iso_to_local"
]

# --------------------------------------------------------------------------- #

def datetime_epoch():
    epoch = "1970-01-01T00:00:00.000000"
    return datetime(1970, 1, 1), epoch

def datetime_now_utc_iso():
    return datetime.now(tz=timezone.utc).isoformat()

def epoch_to_utc_iso(seconds_since_epoch):
    utc = datetime.fromtimestamp(seconds_since_epoch, tz=timezone.utc)
    return utc.isoformat()

def local_iso_now():
  return utc_iso_to_local(datetime_now_utc_iso())

def utc_iso_since_epoch(datetime_utc_iso):
    datetime_utc = utc_iso_to_datetime(datetime_utc_iso)
    return (datetime_utc - datetime_epoch()[0]).total_seconds()

def utc_iso_to_datetime(datetime_utc_iso):
#   datetime_utc = date.fromisoformat(datetime_utc_iso)  # should work :(
    if len(datetime_utc_iso) == 25:
        strp_isoformat = "%Y-%m-%dT%H:%M:%S%z"
    else:
        strp_isoformat = "%Y-%m-%dT%H:%M:%S.%f%z"
    datetime_utc = datetime.strptime(datetime_utc_iso, strp_isoformat)
    return datetime_utc

def utc_iso_to_local(datetime_utc_iso):
    datetime_utc = utc_iso_to_datetime(datetime_utc_iso)
    datetime_local = datetime_utc.replace(
        tzinfo=timezone.utc).astimezone(tz=None)
    return datetime_local.isoformat().replace("T", " ")[:19]

# --------------------------------------------------------------------------- #



# isort: skip_file

from .assessments import *
from .aws_s3 import *
from .base_models import *
from .capabilities import *
from .data_files import *
from .entity_avro_schema import *
from .evaluation import *
from .gql_client import HLClient, S3Creds  # fmt: skip
from .io import *  # fmt: skip
from .object_classes import *
from .presigned_url import *
from .tasks import *
from .training_config import *
from .training_runs import *
from .workflow import *
from .ingestion_setup import *
from .json_tools import *
from .agents import *
from .utilities import *



from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID, uuid4

import numpy as np
import shapely.geometry as geom
from PIL.Image import Image
from pydantic import BaseModel, ConfigDict, Field, field_validator

from highlighter.core import GQLBaseModel
from highlighter.core.utilities import stringify_if_not_null

from ...core import (
    OBJECT_CLASS_ATTRIBUTE_UUID,
    polygon_from_mask,
)
from .base_models import EAVT
from .datum_source import DatumSource
from .observation import Observation
from .validated_set import ValidatedSet

_sentinel = object()


class Annotation(GQLBaseModel):
    id: UUID = Field(..., default_factory=uuid4)
    _entity: Optional[Any] = None  # Optional[Entity], but Entity is not imported to avoid circular imports

    location: Optional[Union[geom.Polygon, geom.MultiPolygon, geom.LineString, geom.Point]] = None
    observations: ValidatedSet
    track_id: Optional[UUID] = None
    correlation_id: Optional[UUID] = None

    _data_sample: Optional[Any] = (
        None  # Optional[DataSample], but DataSample is not imported to avoid circular imports
    )

    # DEPRECATED: Use self.data_sample instead of self.{data_file_id, datum_source.frame_id, occurred_at}
    data_file_id: Optional[UUID] = None
    # Frame ID in datum_source
    datum_source: DatumSource = Field(..., default_factory=lambda: DatumSource(confidence=1))
    occurred_at: Optional[datetime] = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True, extra="ignore"
    )  # Required for shapely geometry types

    def __init__(self, **kwargs):
        observations = ValidatedSet()
        if "observations" in kwargs and kwargs["observations"] is not None:
            for obs in kwargs["observations"]:
                observations.append(obs)
        kwargs["observations"] = observations
        super().__init__(**kwargs)
        if "entity" in kwargs:
            self.entity = kwargs["entity"]  # Trigger setter
        # Validate observations passed to constructor
        self.observations.validator = self.validate_add_observation
        for observation in self.observations:
            self.validate_add_observation(observation)
        if "data_sample" in kwargs and kwargs["data_sample"] is not None:
            self.data_sample = kwargs["data_sample"]  # Trigger setter

    @property
    def entity(self) -> Optional["Entity"]:
        return self._entity

    @entity.setter
    def entity(self, new_entity: Optional["Entity"]):
        from .entity import Entity  # Import here to avoid circular imports

        if not isinstance(new_entity, Optional[Entity]):
            raise TypeError(f"Expected Entity, got '{type(new_entity).__qualname__}'")

        # Check if update is allowed
        if new_entity is not None and self.entity is not None and new_entity != self.entity:
            raise ValueError(
                f"Annotation is still associated with entity {self.entity.id}. "
                "Un-associate via `self.entity = None` before "
                "associating with a different entity."
            )

        # Do the update
        if self.entity is not None:
            self.entity.annotations.remove(self)
        if new_entity is not None and self not in new_entity.annotations:
            new_entity.annotations.add(self)
        self._entity = new_entity

    @property
    def data_sample(self) -> Optional["DataSample"]:
        return self._data_sample

    @data_sample.setter
    def data_sample(self, new_data_sample: Optional["DataSample"]):
        from ...core.data_models.data_sample import (
            DataSample,  # Import here to avoid circular imports
        )

        if new_data_sample is not None and not isinstance(new_data_sample, DataSample):
            raise TypeError(f"Expected DataSample, got '{type(new_data_sample).__qualname__}'")

        self._data_sample = new_data_sample
        if new_data_sample is not None:
            # Data file ID, occurred-at and frame ID are redundant with the associated data sample's values
            self.data_file_id = new_data_sample.data_file_id
            self.occurred_at = new_data_sample.recorded_at
            self.datum_source.frame_id = new_data_sample.media_frame_index
            for obs in self.observations:
                obs._update_from_annotation()

    def validate_add_observation(self, new_observation: Observation):
        if not isinstance(new_observation, Observation):
            raise TypeError(f"Expected Observation, got {type(new_observation).__qualname__}")

        # Check if update is allowed
        if new_observation._global_observation_entity is not None:
            raise ValueError(
                "Cannot add observation that references a different entity. "
                f"Un-associate the observation from entity {new_observation.entity.id} "
                f"via `observation.entity = None` before adding it to annotation {self.id}"
            )
        if new_observation._annotation is not None and new_observation._annotation != self:
            raise ValueError(
                "Cannot add observation that references a different annotation. "
                f"Un-associate the observation from annotation {new_observation._annotation.id} "
                f"via `observation.annotation = None` before adding it to annotation {self.id}"
            )

        # Do associated updates
        new_observation._annotation = self
        # _update_from_annotation is idempotent.
        # Doing `ann.observations.add(obs)` will lead to it being called
        # once here, but doing `obs.annotation = ann` it will be called once in the
        # setter and then again here.
        new_observation._update_from_annotation()

    def get_or_create_entity(self) -> "Entity":
        from .entity import Entity

        if self.entity is not None:
            return self.entity
        entity = Entity()
        entity.annotations.add(self)
        return entity

    @classmethod
    def from_points(
        cls,
        points: List[Tuple[int, int]],
        confidence: float,
        data_sample: Optional[Any] = None,
        data_file_id: Optional[UUID] = None,
        observations: Optional[List[Observation]] = None,
        track_id: Optional[UUID] = None,
        correlation_id: Optional[UUID] = None,
        # Optional DatumSource fields
        frame_id: Optional[int] = None,
        host_id: Optional[str] = None,
        pipeline_element_name: Optional[str] = None,
        occurred_at: Optional[datetime] = None,
    ):
        datum_source = DatumSource(
            confidence=confidence,
            frame_id=frame_id,
            host_id=host_id,
            pipeline_element_name=pipeline_element_name,
        )

        anno = cls(
            location=geom.Polygon(points),
            observations=observations,
            track_id=track_id,
            correlation_id=correlation_id,
            data_sample=data_sample,
            data_file_id=data_file_id,
            datum_source=datum_source,
            occurred_at=occurred_at,
        )
        return anno

    @classmethod
    def from_left_top_right_bottom_box(
        cls,
        box: Tuple[int, int, int, int],
        confidence: float,
        data_sample: Optional[Any] = None,
        data_file_id: Optional[UUID] = None,
        observations: Optional[List[Observation]] = None,
        track_id: Optional[UUID] = None,
        correlation_id: Optional[UUID] = None,
        # Optional DatumSource fields
        frame_id: Optional[int] = None,
        host_id: Optional[str] = None,
        pipeline_element_name: Optional[str] = None,
        occurred_at: Optional[datetime] = None,
    ):
        x0, y0, x1, y1 = box
        return cls.from_points(
            [(x0, y0), (x1, y0), (x1, y1), (x0, y1), (x0, y0)],
            confidence,
            data_sample=data_sample,
            data_file_id=data_file_id,
            observations=observations,
            track_id=track_id,
            correlation_id=correlation_id,
            frame_id=frame_id,
            host_id=host_id,
            pipeline_element_name=pipeline_element_name,
            occurred_at=occurred_at,
        )

    @classmethod
    def from_mask(
        cls,
        mask: np.array,
        confidence: float,
        data_sample: Optional[Any] = None,
        data_file_id: Optional[UUID] = None,
        observations: Optional[List[Observation]] = None,
        track_id: Optional[UUID] = None,
        correlation_id: Optional[UUID] = None,
        # Optional DatumSource fields
        frame_id: Optional[int] = None,
        host_id: Optional[str] = None,
        pipeline_element_name: Optional[str] = None,
        occurred_at: Optional[datetime] = None,
        **polygon_from_mask_kwargs,
    ):

        points = polygon_from_mask(mask, **polygon_from_mask_kwargs)
        points = [(x, y) for x, y in zip(points[:-1:2], points[1::2])]
        # Shapely expects a closed
        points.append(points[0])
        # breakpoint()

        return cls.from_points(
            points,
            confidence,
            data_sample=data_sample,
            data_file_id=data_file_id,
            observations=observations,
            track_id=track_id,
            correlation_id=correlation_id,
            frame_id=frame_id,
            host_id=host_id,
            pipeline_element_name=pipeline_element_name,
            occurred_at=occurred_at,
        )

    def observations_where(
        self, *, attribute_id: UUID | None = None, value: Any = _sentinel
    ) -> List[Observation]:
        obs_where = []
        for o in self.observations:
            if attribute_id is not None and o.attribute_id == attribute_id:
                if value is _sentinel or o.value == value:
                    obs_where.append(o)
        return obs_where

    def crop_image(self, crop_args: Optional["CropArgs"] = None) -> Union[np.ndarray, Image]:
        from ...datasets.cropping import crop_rect_from_poly

        if self.data_sample is None:
            raise ValueError("Annotation must be associated with a DataSample to be cropped")
        if self.location is None:
            raise ValueError("Cannot crop an Annotation when location is None")
        if not isinstance(self.location, geom.Polygon):
            raise ValueError(f"Cannot crop an Annotation when location is {type(self.location)}")
        if "image" not in self.data_sample.content_type:
            raise ValueError(
                f"Cannot crop data_file with content_type '{self.data_sample.content_type}', must be 'image'"
            )
        if not isinstance(self.data_sample.content, (np.ndarray, Image)):
            raise ValueError(
                f"Cannot crop data_file with content '{type(self.data_sample.content).__qualname__}', must be (PIL.Image|np.array)"
            )

        return crop_rect_from_poly(self.data_sample.content, self.location, crop_args)

    def to_deprecated_pixel_location_eavt(self) -> EAVT:
        if self.entity is None:
            raise ValueError("Entity must exist to convert Annotation to deprecated EAVT")
        return EAVT.make_pixel_location_eavt(
            entity_id=self.entity.id,
            location_points=self.location,
            confidence=self.datum_source.confidence,
            time=datetime.now(timezone.utc),
            frame_id=self.datum_source.frame_id,
        )

    @field_validator("location")
    @classmethod
    def validate_geometry(cls, v):
        if v is not None:
            assert v.is_valid, f"Invalid Geometry: {v}"
        return v

    def serialize(self):
        return {
            "id": str(self.id),
            "entity_id": str(self.entity.id) if self.entity else None,
            "location": self.location.wkt if self.location is not None else None,
            "observations": [o.serialize() for o in self.observations],
            "track_id": str(self.track_id),
            "correlation_id": str(self.correlation_id),
            "data_file_id": str(self.data_file_id),
            "datum_source": self.datum_source.serialize(),
        }

    def to_json(self):
        data = self.model_dump()
        data["datum_source"] = self.datum_source.model_dump()
        data["id"] = stringify_if_not_null(data["id"])
        data["entity_id"] = stringify_if_not_null(data.get("entity", {}).get("id"))
        data["location"] = data["location"].wkt if data["location"] is not None else None
        data["observations"] = [d.to_json() for d in data["observations"]]

        data["track_id"] = stringify_if_not_null(data["track_id"])
        data["data_file_id"] = stringify_if_not_null(data["data_file_id"])
        data["correlation_id"] = stringify_if_not_null(data["correlation_id"])
        data["occurred_at"] = self.occurred_at.isoformat() if self.occurred_at is not None else None
        return data

    def gql_dict(self) -> Dict:
        try:
            object_class_observation = [
                observation
                for observation in self.observations
                if observation.attribute_id == OBJECT_CLASS_ATTRIBUTE_UUID
            ][-1]
        except IndexError:
            raise ValueError(
                "Annotation must have an object-class observation in order to submit to Highlighter"
            )

        if isinstance(self.location, geom.Polygon):
            data_type = "polygon"
        elif isinstance(self.location, geom.LineString):
            data_type = "line"
        elif isinstance(self.location, geom.Point):
            data_type = "point"
        else:
            data_type = "polygon"
        result = {
            "objectClassUuid": stringify_if_not_null(object_class_observation.value),
            "location": self.location.wkt if self.location is not None else None,
            "confidence": self.datum_source.confidence,
            "dataType": data_type,
            "correlationId": stringify_if_not_null(self.correlation_id),
            "frameId": self.datum_source.frame_id,
            "trackId": stringify_if_not_null(self.track_id),
            "entityId": str(self.entity.id) if self.entity else None,
            "dataFileId": stringify_if_not_null(self.data_file_id),
            "uuid": stringify_if_not_null(self.id),
        }
        return result

    def to_deprecated_observations(self) -> List[Observation]:
        """
        Convert to the deprecated "set of observations" representation
        where pixel locations are represented as observations rather than annotations
        """
        observations = list(self.observations)
        if self.location is not None:
            observation = Observation.make_pixel_location_observation(
                value=self.location,
                confidence=self.datum_source.confidence,
                occurred_at=self.occurred_at,
                frame_id=self.datum_source.frame_id,
            )
            observation.annotation = self
            observations.append(observation)
        return observations



# Usage: File
# ~~~~~~~~~~~
# aiko_pipeline create pipelines/image_pipeline_0.json -s 1 -sr -ll debug
#
# aiko_pipeline create pipelines/image_pipeline_0.json -s 1 -p rate 1.0
#
# aiko_pipeline create pipelines/image_pipeline_0.json -s 1  \
#   -p ImageReadFile.data_batch_size 8
#
# aiko_pipeline create pipelines/image_pipeline_0.json -s 1  \
#   -p ImageReadFile.data_sources file://data_in/in_{}.jpeg
#
# aiko_pipeline create pipelines/image_pipeline_0.json -s 1  \
#   -p ImageWriteFile.path "file://data_out/out_{:02d}.jpeg"
#
# aiko_pipeline create pipelines/image_pipeline_0.json -s 1  \
#   -p ImageReadFile.data_sources file://data_in/in_00.jpeg  \
#   -p ImageResize.resolution 320x240                        \
#   -p ImageWriteFile.data_targets file://data_out/out_00.jpeg
#
# aiko_pipeline create pipelines/image_pipeline_1.json -s 1
#
# Usage: ZMQ
# ~~~~~~~~~~
# aiko_pipeline create pipelines/image_zmq_pipeline_0.json -s 1 -sr  \
#            -ll debug -gt 10
# aiko_pipeline create pipelines/image_zmq_pipeline_0.json -s 1 -sr  \
#            -p ImageReadZMQ.data_sources zmq://0.0.0.0:6502
#
# aiko_pipeline create pipelines/image_zmq_pipeline_1.json -s 1 -sr  \
#            -ll debug                                               \
#            -p ImageReadFile.rate 2.0                               \
#            -p ImageWriteZMQ.data_targets zmq://192.168.0.1:6502
#
# To Do
# ~~~~~
# - Support for "media type" encoding details for "image"
#   - Consider additional encoding information in out-of-band image records ?
#     - "frame_id" and/or "image:length:content", "image/jpeg:length:content" ?
#
# - Consolidate multiple "media_pipeline_?.json" files into Graph Path(s) !
# - ImageResize: {"scale": "1/3"} or {"scale": 2} or {"resolution": 640}
#
# - Refactor optional module import into common function (see video_io.py)
# - Provide checks around "cv2" and "numpy" usage to be optional
#
# - Consider what causes Stream to be closed, e.g single frame processed ?
#   - When all [data_sources] produced and not the default stream ?
#
# - Add Stream "parameter" for optional output Image numpy format
#
# - Document "DataSource" and "DataTarget" ...
#   - ImageReadFile:  start_stream() --> create_frames() --> process_frame()
#                     Therefore pass parameters as required
#   - ImageWriteFile: Mustn't use create_frame()
#                     Therefore get parameters in process_frame()
#
# - ImageReadFile should accept a list of DataSources type ...
#   - URL: "file://" and media_type: "jpeg", "png", etc
#
# - Metrics: Determine what to metrics to capture, e.g frame rates ?

from PIL import Image
import io
from typing import Tuple
import zlib

import aiko_services as aiko

__all__ = [
    "bytes_to_image", "image_to_bytes",
    "convert_image", "convert_images",
    "convert_image_to_numpy", "convert_image_to_pil",
    "ImageConvert", "ImageOutput", "ImageOverlay", "ImageOverlayFilter",
    "ImageReadFile", "ImageReadZMQ", "ImageResize", "ImageSquareCenterCrop",
    "ImageWriteFile", "ImageWriteZMQ"
]

_LOGGER = aiko.process.logger(__name__)

_CV2_IMPORTED = False
try:
    import cv2
    _CV2_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "image_io.py: Couldn't import cv2 module"
#   print(f"WARNING: {diagnostic}")
#   _LOGGER.warning(diagnostic)
#   raise ModuleNotFoundError(
#       'opencv-python package not installed.  '
#       'Install aiko_services with --extras "opencv" '
#       'or install opencv-python manually to use the "image_io" module')

_NUMPY_IMPORTED = False
try:
    import numpy as np
    _NUMPY_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "image_io.py: Couldn't import numpy module"
#   print(f"WARNING: {diagnostic}")
#   _LOGGER.warning(diagnostic)

# --------------------------------------------------------------------------- #
# str: "JPEG" or "PNG"

def image_to_bytes(image: Image.Image, format: str="JPEG") -> bytes:
    with io.BytesIO() as image_bytes_io:
        image = convert_image_to_pil(image)
        if not image:
            raise ValueError(
                f"image_to_bytes(): Unknown image type: {type(image)}")
        image.save(image_bytes_io, format=format)
        return image_bytes_io.getvalue()

def bytes_to_image(image_bytes: bytes) -> Image.Image:
    image_bytes_io = io.BytesIO(image_bytes)
    image = Image.open(image_bytes_io)
    image.load()
    image_bytes_io.close()
    return image

def convert_image_to_numpy(image):
    if not isinstance(image, np.ndarray):
        if isinstance(image, Image.Image):
            image = np.array(image)  # RGB
        else:
            image = None
    return image

def convert_image_to_pil(image):
    if not isinstance(image, Image.Image):
        if isinstance(image, np.ndarray):  # TODO: Check NUMPY_IMPORTED
            image = Image.fromarray(image.astype("uint8"), "RGB")
        else:
            image = None
    return image

convert_image_handlers = {
    "numpy": convert_image_to_numpy,
    "pil": convert_image_to_pil
}

def convert_image(image, media_type):
    if media_type not in convert_image_handlers:
        raise ValueError(
            f"image_io:convert_image(): Unknown media type: {media_type}")
    return convert_image_handlers[media_type](image)

def convert_images(images, media_type):
    images_converted = []
    for image in images:
        images_converted.append(convert_image(image, media_type))
    return(images_converted)

# --------------------------------------------------------------------------- #

class ImageConvert(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_convert:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        media_type, _ = self.get_parameter("media_type", None)
        if media_type:
            images = convert_images(images, media_type)
        return aiko.StreamEvent.OKAY, {"images": images}

# --------------------------------------------------------------------------- #
# Useful for Pipeline output that should be all of the images processed

class ImageOutput(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_output:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        return aiko.StreamEvent.OKAY, {"images": images}

# --------------------------------------------------------------------------- #
# TODO: Change color (text, boxes face / YOLO), font, line thickness
# TODO: Display camera name/path, date/time, FPS
# TODO: Display specified elipses, lines, mask, rectangles, polygons, text
# TODO: Display metrics
# TODO: Display pose stick figures

class ImageOverlay(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("image_overlay:0")
        context.call_init(self, "PipelineElement", context)
        self.color = (0, 255, 255)
        self.font = cv2.FONT_HERSHEY_SIMPLEX
        self.font_scale = 0.75
        self.thickness = 2
        self.threshold = 0.0

    def process_frame(self, stream, images, overlay)  \
        -> Tuple[aiko.StreamEvent, dict]:

        images_overlaid = []
        for image in images:
            image_width = image.shape[1]
            grayscale = len(image.shape) == 2
            if grayscale:
                image_bgr = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
            else:
                image_bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

            if "rectangles" in overlay:
                rectangles = overlay["rectangles"]
                if "objects" in overlay:
                    objects = overlay["objects"]
                else:
                    objects = [{}] * len(rectangles)

                items = zip(objects, rectangles)
                for item in items:
                    object, rectangle = item
                    name = object.get("name", None)
                    confidence = object.get("confidence", 1.0)
                    if confidence >= self.threshold:
                        x = int(rectangle["x"])
                        y = int(rectangle["y"])
                        w = int(rectangle["w"])
                        h = int(rectangle["h"])
                        top_left = (x, y)
                        bottom_right = (x + w, y + h)
                        if name:
                            text = f"{name}: {confidence:0.2f}"
                            width_height, _ = cv2.getTextSize(text,
                                self.font, self.font_scale, self.thickness)
                            tw, th = width_height
                            tx = x
                            if y > th * 1.5:
                                ty = int(y - th / 2 - 2)
                            else:
                                ty = y + h + th + 2
                            if x + tw > image_width:
                                tx = x - tw
                                ty = y + th
                            image_bgr = cv2.putText(image_bgr, text, (tx, ty),
                                self.font, self.font_scale, self.color,
                                self.thickness, cv2.LINE_AA)
                        image_bgr = cv2.rectangle(image_bgr,
                            top_left, bottom_right, self.color, self.thickness)

            if grayscale:
                image_overlaid = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2GRAY)
            else:
                image_overlaid = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2RGB)
            images_overlaid.append(image_overlaid)

        return aiko.StreamEvent.OKAY, {"images": images_overlaid}

# --------------------------------------------------------------------------- #

class ImageOverlayFilter(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("image_overlay_filter:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, overlay)  \
        -> Tuple[aiko.StreamEvent, dict]:

        deny, _ = self.get_parameter("deny", [])
        threshold, _ = self.get_parameter("threshold", 0.0)

        overlay_filtered = {"objects": [], "rectangles": []}
        objects_filtered = overlay_filtered["objects"]
        rectangles_filtered = overlay_filtered["rectangles"]

        if "rectangles" in overlay:
            rectangles = overlay["rectangles"]
            if "objects" in overlay:
                objects = overlay["objects"]
            else:
                objects = [{}] * len(rectangles)

            items = zip(objects, rectangles)
            for item in items:
                object, rectangle = item
                name = object.get("name", None)
                confidence = object.get("confidence", 1.0)

                if name not in deny and confidence >= threshold:
                    objects_filtered.append(object)
                    rectangles_filtered.append(rectangle)

        return aiko.StreamEvent.OKAY, {"overlay": overlay_filtered}

# --------------------------------------------------------------------------- #
# ImageReadFile is a DataSource which supports ...
# - Individual image files
# - Directory of image files with an optional filename filter
# - TODO: Archive (tgz, zip) of image files with an optional filename filter
#
# parameter: "data_sources" is the read file path, format variable: "frame_id"
#
# Note: Only supports Streams with "data_sources" parameter

class ImageReadFile(aiko.DataSource):  # PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_read_file:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, paths) -> Tuple[aiko.StreamEvent, dict]:
        stream.variables["timestamps"] = [stream.frame_id * (1 / 25)]
        images = []
        media_type, _ = self.get_parameter("media_type", None)
        for path in paths:
            try:
                image = Image.open(path)
                if media_type:
                    image = convert_image(image, media_type)
                images.append(image)
                self.logger.debug(f"{self.my_id()}: {path} {image.size}")
            except Exception as exception:
                diagnostic = f"Error loading image: {exception}"
                return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        return aiko.StreamEvent.OKAY, {"images": images}

# --------------------------------------------------------------------------- #
# ImageReadZMQ is a DataSource which supports ...
# - ImageWriteZMQ(DataTarget) ZMQ client --> ImageReadZMQ(DataSource) ZMQ server
#   - Individual image records produced by ZMQ client and consumed by ZMQ server
#
# parameter: "data_sources" is the ZMQ server bind details (scheme_zmq.py)
#            "media_type" is either "numpy" or "pil"
#            "compressed" (boolean) indicated whether to decompress the payload
#
# Note: Only supports Streams with "data_sources" parameter

class ImageReadZMQ(aiko.DataSource):  # PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_read_zmq:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, records) -> Tuple[aiko.StreamEvent, dict]:
        stream.variables["timestamps"] = [stream.frame_id * (1 / 25)]
        images = []
        media_type, _ = self.get_parameter("media_type", None)
        if media_type:
            tokens = media_type.split("/")
            media_type = tokens[0] if len(tokens) == 1 else tokens[1]

        compressed, _ = self.get_parameter("compressed", False)

        for record in records:
            record = zlib.decompress(record) if compressed else record
            image = bytes_to_image(record)
            self.logger.debug(f"{self.my_id()}: {len(record)} --> {image.size}")
    #       if image.startswith("image:"):  # TODO: "image:length:content" ?
    #           tokens = image.split(":")
    #           image = tokens[2:][0]      # just the "content"
            if media_type:
                image = convert_image(image, media_type)
            images.append(image)
        return aiko.StreamEvent.OKAY, {"images": images}

# --------------------------------------------------------------------------- #
# TODO: Combine ImageResize and ImageSquareCenterCrop PipelineElements
# TODO: Add logic for using different backends (opencv,...) or
#           different input types (np.ndarray, ...)
#
# TODO: cv2.resize(image, dimensions, interpolation=cv2.INTER_CUBIC) ?
#       Shrink: INTER_AREA, enlarge: INTER_CUBIC (slow) or INTER_LINEAR (fast)
#       parameter: "speed": "fast" --> INTER_AREA or INTER_CUBIC
#       parameter: "speed": "slow" --> INTER_LINEAR
#       https://gist.github.com/georgeblck/e3e0274d725c858ba98b1c36c14e2835

class ImageResize(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_resize:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        resolution, found = self.get_parameter("resolution")
        if not found:
            return aiko.StreamEvent.OKAY, {"images": images}

        width, height = resolution.split("x")
        self.logger.debug(f"{self.my_id()}: resolution: {width}x{height}")

        images_resized = []
        for image in images:
            if isinstance(image, np.ndarray):  # TODO: Check CV2_IMPORTED
                image_resized = cv2.resize(image, (int(width), int(height)))
            else:
                image_resized = image.resize((int(width), int(height)))
            images_resized.append(image_resized)
        return aiko.StreamEvent.OKAY, {"images": images_resized}

# --------------------------------------------------------------------------- #
# TODO: Combine ImageResize and ImageSquareCenterCrop PipelineElements
#       ImageResize should be able to handle both Pillow and NumPy array
#       ImageResize should have a "square_centre_crop=boolean" parameter

def square_center_crop(image):  # Either Pillow or NumPy array Image
    if isinstance(image, Image.Image):
        width, height = image.size
        is_pil = True
    elif isinstance(image, np.ndarray):
        height, width = image.shape[:2]
        is_pil = False
    else:
        raise TypeError(
            f"square_center_crop() requires PIL.Image or np.ndarray, "
            f"but received type {type(image)}"
        )

    side = min(width, height)  # Maintain shortest side
    left = (width - side) // 2
    top  = (height - side) // 2

    if is_pil:
        return image.crop((left, top, left + side, top + side))
    else:
        return image[top : top + side, left : left + side, ...]

class ImageSquareCenterCrop(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_square_center_crop:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        images_cropped = []
        for image in images:
            images_cropped.append(square_center_crop(image))
        return aiko.StreamEvent.OKAY, {"images": images_cropped}

# --------------------------------------------------------------------------- #
# ImageWriteFile is a DataTarget that writes images to files
#
# parameter: "data_targets" is the write file path, format variable: "frame_id"
#
# Note: Only supports Streams with "data_targets" parameter

class ImageWriteFile(aiko.DataTarget):  # PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_write_file:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        for image in images:
            path = stream.variables["target_path"]
            if stream.variables["target_path_template"]:
                path = path.format(stream.variables["target_file_id"])
                stream.variables["target_file_id"] += 1
            self.logger.debug(f"{self.my_id()}: {path}")

            image = convert_image_to_pil(image)
            if not image:
                diagnostic = "UNKNOWN IMAGE TYPE"  # TODO: FIX ME !
                return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}
            try:
                image.save(path)
            except Exception as exception:
                diagnostic = f"Error saving image: {exception}"
                return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #
# ImageWriteZMQ is a DataTarget which supports ...
# - ImageWriteZMQ(DataTarget) ZMQ client --> ImageReadZMQ(DataSource) ZMQ server
#   - Individual image records produced by ZMQ client and consumed by ZMQ server
#
# parameter: "data_targets" is the ZMQ connect details (scheme_zmq.py)
#            "compressed" (boolean) indicated whether to compress the payload
#
# Note: Only supports Streams with "data_targets" parameter

class ImageWriteZMQ(aiko.DataTarget):  # PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_write_zmq:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        media_type = "image"                            # TODO: "image/zip" ?
        compressed, _ = self.get_parameter("compressed", False)

        for image in images:
     #      image = f"{media_type}:{len(image)}:{image}"  # "image:length:content" ?
            record = image_to_bytes(image)
            record = zlib.compress(record) if compressed else record
            self.logger.debug(f"{self.my_id()}: {image.size} --> {len(record)}")
            stream.variables["target_zmq_socket"].send(record)

        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #



from __future__ import annotations

from typing import Annotated, Literal, Union

import numpy as np
from numpy.typing import NDArray
from pydantic import BaseModel, Field

__all__ = [
    "OutputTransform",
    "OutputTransformUnion",
    "UnityOutputTransform",
    "LogisticOutputTransform",
    "ScaleOutputTransform",
]


class OutputTransform(BaseModel):
    def __call__(self, x: Union[float, NDArray[np.float_]]) -> Union[float, NDArray[np.float_]]:
        raise NotImplementedError()


class UnityOutputTransform(OutputTransform):
    type: Literal["UnityOutputTransform"] = "UnityOutputTransform"

    def __call__(self, x: Union[float, NDArray[np.float_]]) -> Union[float, NDArray[np.float_]]:
        return x


class LogisticOutputTransform(OutputTransform):
    type: Literal["LogisticOutputTransform"] = "LogisticOutputTransform"
    steepness: float
    centre: float
    scale: float

    def __call__(self, x: Union[float, NDArray[np.float_]]) -> Union[float, NDArray[np.float_]]:
        return self.scale * (1.0 / (1.0 + np.exp(-self.steepness * (x - self.centre))))


class ScaleOutputTransform(OutputTransform):
    type: Literal["ScaleOutputTransform"] = "ScaleOutputTransform"
    scale_factor: float = 1.0

    def __call__(self, x: Union[float, NDArray[np.float_]]) -> Union[float, NDArray[np.float_]]:
        return x * self.scale_factor


OutputTransformUnion = Annotated[
    Union[UnityOutputTransform, LogisticOutputTransform, ScaleOutputTransform],
    Field(discriminator="type"),  # discriminator on the *field*, not on the models
]



# Usage
# ~~~~~
# - aiko_pipeline create pipelines/factorial_pipeline.json -ll debug_all  \
#                                                          -fd "()"
#
# To Do
# ~~~~~
# - "PipelineElementLoop": Log debug for "define", "condition", "expression"

from typing import Tuple

import aiko_services as aiko
from aiko_services.elements.utilities import evaluate_condition, evaluate_define
from aiko_services.main.utilities import parse

__all__ = ["Loop"]

# --------------------------------------------------------------------------- #

class Loop(aiko.PipelineElementLoop):
    def __init__(self, context):
        context.set_protocol("loop:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream) -> Tuple[aiko.StreamEvent, dict]:
        self.logger.debug(f"{self.my_id()}")
        swag = stream.frames[stream.frame_id].swag

        if stream.variables.get("loop_boundary", None) is None:
            boundary, _ = self.get_parameter("boundary", "")
            stream.variables["loop_boundary"] = boundary

            expression, found = self.get_parameter("define")
            if found:
                expression = parse(expression, car_cdr=False)
                evaluate_define(expression, swag)

        condition, found = self.get_parameter("condition")
        if not found:
            diagnostic = 'Must provide a "condition" parameter'
            return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        condition = parse(condition, car_cdr=False)
        condition = evaluate_condition(condition, swag)
        if condition:
            expression, found = self.get_parameter("expression")
            if found:
                expression = parse(expression, car_cdr=False)
                evaluate_define(expression, swag)
            event = aiko.StreamEvent.OKAY
        else:
            self.logger.debug(f"End")
            event = aiko.StreamEvent.LOOP_END

        return event, {}

# --------------------------------------------------------------------------- #



import json
import logging
from pathlib import Path
from typing import Dict, List

import click
import fastavro as avro
import yaml

from ..client import (
    ENTITY_AVRO_SCHEMA,
    HLJSONEncoder,
    PageInfo,
    SubmissionType,
    create_assessment_with_avro_file,
    get_latest_assessments_gen,
    upload_file_to_s3,
)
from ..core import GQLBaseModel, paginate
from ..datasets import Dataset
from ..datasets.formats.coco.reader import CocoReader
from ..datasets.formats.highlighter.writer import (
    CreateSubmissionPayload,
    HighlighterAssessmentsWriter,
)

ASSESSMENT_FIELDS = list(SubmissionType.__annotations__.keys())


class AnnotationsAttributesParamType(click.ParamType):
    name = "dict"

    def convert(self, value, param, ctx):
        try:
            return json.loads(value)
        except json.JSONDecodeError as e:
            self.fail(f"Invalid JSON: {e}", param, ctx)


class EAVTAttributesParamType(click.ParamType):
    name = "dict"

    def convert(self, value, param, ctx):
        try:
            return json.loads(value)
        except json.JSONDecodeError as e:
            self.fail(f"Invalid JSON: {e}", param, ctx)


@click.group("assessment")
@click.pass_context
def assessment_group(ctx):
    pass


@assessment_group.command("create")
@click.option(
    "-w",
    "--workflow-id",
    type=int,
    required=True,
    help="Destination process",
)
@click.option(
    "-u",
    "--user-id",
    type=int,
    required=False,
    help="User ID to load submission against",
)
@click.option(
    "-i",
    "--data-file-id",
    type=click.UUID,
    required=True,
    help="File UUID",
)
@click.option(
    "-s",
    "--status",
    type=str,
    required=True,
    help="Assessment status for assessment",
)
@click.option(
    "-m",
    "--capability-id",
    type=int,
    required=False,
    help="Assessment status for assessment",
)
@click.option(
    "-c",
    "--matched-image-id",
    type=int,
    required=False,
    help="Matched image ID for assessment",
)
@click.option(
    "-t",
    "--training-run-id",
    type=int,
    required=False,
    help="Training run ID for assessment",
)
@click.option(
    "-q",
    "--step-id",
    type=int,
    required=False,
    help="Image queue ID for assessment",
)
@click.option(
    "-d",
    "--data-source-id",
    type=int,
    required=False,
    help="Data source ID for assessment",
)
@click.option(
    "-a",
    "--started-at",
    type=str,
    required=False,
    help="The datetime this assessment was started",
)
@click.option(
    "-f",
    "--flag-reason",
    type=str,
    required=False,
    help="The reason this assessments is flagged",
)
@click.option(
    "-b",
    "--background-info-layer-file-data",
    type=str,
    required=False,
    help="Background information layer file data",
)
@click.option(
    "-n",
    "--annotations-attributes",
    type=AnnotationsAttributesParamType(),
    required=False,
    help="Annotations attributes",
)
@click.option(
    "-v",
    "--eavt-attributes",
    type=EAVTAttributesParamType(),
    required=False,
    help="EAVT attributes",
)
@click.option(
    "-k",
    "--task-id",
    type=str,
    required=False,
    help="Associated task ID",
)
@click.pass_context
def create(
    ctx,
    workflow_id,
    user_id,
    data_file_id,
    status,
    capability_id,
    matched_image_id,
    training_run_id,
    step_id,
    data_source_id,
    started_at,
    flag_reason,
    background_info_layer_file_data,
    annotations_attributes,
    eavt_attributes,
    task_id,
):
    client = ctx.obj["client"]

    result = client.createSubmission(
        return_type=CreateSubmissionPayload,
        workflowId=workflow_id,
        userId=user_id,
        dataFileIds=[data_file_id],
        status=status,
        modelId=capability_id,
        trainingRunId=training_run_id,
        imageQueueId=step_id,
        dataSourceId=data_source_id,
        startedAt=started_at,
        flagReason=flag_reason,
        backgroundInfoLayerFileData=background_info_layer_file_data,
        taskId=task_id,
        annotationsAttributes=annotations_attributes,
        eavtAttributes=eavt_attributes,
    ).gql_dict()

    click.echo(json.dumps(result, cls=HLJSONEncoder))


@assessment_group.command("create-from-dataset")
@click.option(
    "-f",
    "--dataset-format",
    type=click.Choice(["hdf", "coco", "json"]),
    required=True,
    help="The format the data is on disk",
)
@click.option(
    "-d",
    "--data-path",
    type=click.Path(dir_okay=False, exists=True),
    required=True,
    help="Path to data file",
)
@click.option(
    "-w",
    "--workflow-id",
    type=int,
    required=True,
    help="Destination workflow",
)
@click.option(
    "-u",
    "--user-id",
    type=int,
    required=False,
    help="User id to load assessments against",
    default=None,
)
@click.option(
    "-s",
    "--data-source-ids",
    type=int,
    required=False,
    help="Where the existing data_files are in Highlighter",
    default=None,
    multiple=True,
)
@click.option(
    "--fix-invalid-polygons/--no-fix-invalid-polygons",
    type=bool,
    required=False,
    help="Try to convert self-intersecting Polygons to MultiPolygon from 'coco' dataset",
    default=False,
)
@click.pass_context
def create_from_dataset(
    ctx,
    dataset_format,
    data_path,
    workflow_id,
    user_id,
    data_source_ids,
    fix_invalid_polygons,
):
    """Create assessments from an on disk dataset.

    It is expected that data_files have been uploaded already to one of the provided
    --data-source-ids. If you use `hl data_file create` with a coco dataset it will
    produce a directory containing an augmented coco dataset with symlinks mapping
    the Highlighter data_file_id to the original file path. We recommend this workflow
    in most cases.

    """
    client = ctx.obj["client"]
    logger = logging.getLogger(__name__)

    if dataset_format == "hdf":
        if fix_invalid_polygons:
            logger.info(
                "--fix-invalid-polygons only used for --dataset-format 'coco'. It has no effect when using 'hdf'"
            )
        ds = Dataset.read_hdf(path=data_path)
    elif dataset_format == "json":
        if fix_invalid_polygons:
            logger.info(
                "--fix-invalid-polygons only used for --dataset-format 'coco'. It has no effect when using 'json'"
            )
        ds = Dataset.read_json(path=data_path)
    elif dataset_format == "coco":
        reader = CocoReader(
            data_path,
            image_dir=Path(data_path).parent,
            fix_invalid_polygons=fix_invalid_polygons,
        )
        ds = Dataset.load_from_reader(reader)
    else:
        raise ValueError(f"Invalid dataset_format: {dataset_format}")

    # Get data_file ids in data_source
    class File(GQLBaseModel):
        id: int
        original_source_url: str

    class FileConnection(GQLBaseModel):
        page_info: PageInfo
        nodes: List[File]

    if data_source_ids:
        original_source_url_to_id: Dict[str, int] = {
            i.original_source_url: i.id
            for i in paginate(client.fileConnection, FileConnection, dataSourceId=data_source_ids)
        }

        # update data_file_ids
        old_id_filename = ds.data_files_df.loc[:, ["data_file_id", "filename"]].values
        old_id_to_new_id = {
            old_id: original_source_url_to_id[filename] for old_id, filename in old_id_filename
        }
        ds.data_files_df.loc[:, "data_file_id"] = ds.data_files_df.data_file_id.map(old_id_to_new_id)
        ds.annotations_df.loc[:, "data_file_id"] = ds.annotations_df.data_file_id.map(old_id_to_new_id)

    writer = HighlighterAssessmentsWriter(client, workflow_id, user_id=user_id)
    writer.write(ds)


@assessment_group.command("create-from-avro")
@click.option(
    "--avro-file",
    type=click.Path(dir_okay=False, exists=True),
    required=True,
    help="Path to Avro binary data file",
)
@click.option(
    "--file-id",
    type=int,
    required=True,
    help="Highlighter file ID for a video file",
)
@click.option(
    "-w",
    "--workflow-id",
    type=int,
    required=True,
    help="Highlighter Workflow ID for the assessment",
)
@click.option(
    "--data-source-uuid",
    type=str,
    required=False,
)
@click.pass_context
def create_from_avro(ctx, avro_file, file_id, workflow_id, data_source_uuid):
    client = ctx.obj["client"]
    # Check Avro file against schema
    with open(avro_file, "rb") as f:
        list(avro.reader(f, reader_schema=ENTITY_AVRO_SCHEMA))
    file_info = upload_file_to_s3(
        client, avro_file, mimetype="application/octet-stream", data_source_uuid=data_source_uuid
    )
    create_assessment_with_avro_file(client, workflow_id, file_id, file_info)
    print(f"Successfully created assessment on file {file_id} in workflow {workflow_id}")


@assessment_group.command(
    "read",
    help="Read assessments given a query. For simple queries us --query for more "
    "complex queries use --yaml-file. NOTE: !!!  graphQL query keys use camelCase  !!!",
)
@click.option(
    "-q",
    "--query",
    type=str,
    required=False,
    default=None,
    help="simple query yaml string",
)
@click.option(
    "-y",
    "--yaml-file",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    default=None,
    help="yaml file containing a query params that would be a pain to type in the prompt",
)
@click.option(
    "-c",
    "--count",
    type=int,
    required=False,
    default=None,
    help="Only show the first --count results",
)
@click.option(
    "-f",
    "--fields",
    type=str,
    required=False,
    default=None,
    multiple=True,
    help="Name of fields you wish to return, valid fields are: "
    f"{ASSESSMENT_FIELDS}. If blank will return all.",
)
@click.option(
    "--data-file-id",
    type=click.UUID,
    required=False,
    default=None,
    help="Filter assessments by data file UUID",
)
@click.pass_context
def read(ctx, query, yaml_file, count, fields, data_file_id):
    client = ctx.obj["client"]

    if query is not None:
        query_args = yaml.safe_load(query)

    elif yaml_file is not None:
        with open(yaml_file, "r") as f:
            query_args = yaml.safe_load(f)

    else:
        raise ValueError("Expected one of --query or --yaml-file to be set")

    # Add data_file_id to query_args if provided
    if data_file_id:
        query_args["fileId"] = str(data_file_id)

    subs_gen = get_latest_assessments_gen(
        client,
        **query_args,
    )

    if count is None:
        count = float("inf")

    output_list = []
    for i, sub in enumerate(subs_gen, 1):
        sub_dict = sub.gql_dict()

        if fields:
            sub_dict = {f: sub_dict[f] for f in fields}

        output_list.append(sub_dict)

        if i == count:
            break

    click.echo(json.dumps(output_list, cls=HLJSONEncoder))



import logging
import threading
import time
from typing import List, Optional, Union
from uuid import UUID

from ..core import paginate
from .base_models import FileConnection
from .base_models import FilePresigned as FilePresignedType
from .gql_client import HLClient

__all__ = [
    "get_presigned_url",
    "get_presigned_urls",
]


def get_presigned_url(
    client: HLClient,
    id: int,
):
    """Return a single FilePresignedType BaseModel
    for the given file id
    """
    logger = logging.getLogger(__name__)
    current_thread = threading.current_thread()
    thread_id = current_thread.ident
    thread_name = current_thread.name

    logger.debug(f"[Thread {thread_id}/{thread_name}] Calling client.file() for id={id}")
    start = time.perf_counter()
    result = client.file(
        return_type=FilePresignedType,
        id=id,
    )
    elapsed = time.perf_counter() - start
    logger.info(f"[Thread {thread_id}/{thread_name}] client.file() took {elapsed:.3f}s")
    return result


def get_presigned_urls(
    client,
    ids: List[int],
    uuids: List[Union[UUID, str]],
):
    """Return a generator of FilePresignedType GQLBaseModels
    for the given list of file ids
    """
    kwargs = {}
    if ids:
        kwargs["id"] = ids
    if uuids:
        kwargs["uuid"] = uuids

    return paginate(
        client.fileConnection,
        FileConnection,
        **kwargs,
    )



"""Shared discovery infrastructure for CLI commands.

Provides common functionality for mDNS device discovery used by
`hl datasource discover` subcommands.
"""

from __future__ import annotations

from typing import List, Sequence

import click

from highlighter.network.mdns import (
    DEFAULT_DEVICE_KEYWORDS,
    DEFAULT_SERVICE_TYPES,
    DiscoveredDevice,
    DiscoveryConfig,
    discover_devices,
)

DEFAULT_TIMEOUT = 5
DEFAULT_MAX_MAC_WORKERS = 8


def build_discovery_config(
    *,
    timeout: int,
    service_types: Sequence[str],
    keywords: Sequence[str],
    max_mac_workers: int,
) -> DiscoveryConfig:
    """Build DiscoveryConfig from CLI options.

    Args:
        timeout: Discovery timeout in seconds
        service_types: mDNS service types (empty = use defaults)
        keywords: Filter keywords (empty = use defaults)
        max_mac_workers: Max concurrent MAC resolution workers

    Returns:
        Configured DiscoveryConfig instance
    """
    return DiscoveryConfig(
        timeout_seconds=timeout,
        service_types=list(service_types) or list(DEFAULT_SERVICE_TYPES),
        device_keywords=list(keywords) or list(DEFAULT_DEVICE_KEYWORDS),
        max_mac_workers=max_mac_workers,
    )


def run_discovery(
    *,
    timeout: int,
    service_types: Sequence[str],
    keywords: Sequence[str],
    max_mac_workers: int,
    resolve_macs: bool,
) -> List[DiscoveredDevice]:
    """Run mDNS discovery with given options.

    Args:
        timeout: Discovery timeout in seconds
        service_types: mDNS service types to query
        keywords: Keywords to filter services
        max_mac_workers: Max concurrent MAC resolution workers
        resolve_macs: Whether to resolve MAC addresses via ARP

    Returns:
        List of discovered devices
    """
    config = build_discovery_config(
        timeout=timeout,
        service_types=service_types,
        keywords=keywords,
        max_mac_workers=max_mac_workers,
    )
    return discover_devices(config=config, resolve_macs=resolve_macs)


def common_discovery_options(func):
    """Decorator adding common discovery CLI options.

    Adds: --service-type, --keyword, --max-mac-workers, --timeout/-t

    Usage:
        @discover_group.command("list")
        @common_discovery_options
        def list_devices(timeout, service_types, keywords, max_mac_workers):
            ...
    """
    func = click.option(
        "-t",
        "--timeout",
        type=int,
        default=DEFAULT_TIMEOUT,
        show_default=True,
        help="Timeout in seconds for mDNS discovery",
    )(func)
    func = click.option(
        "--max-mac-workers",
        type=int,
        default=DEFAULT_MAX_MAC_WORKERS,
        show_default=True,
        help="Maximum concurrent MAC resolution tasks",
    )(func)
    func = click.option(
        "--keyword",
        "keywords",
        multiple=True,
        help="Filter services by keyword (repeatable, default targets common camera vendors)",
    )(func)
    func = click.option(
        "--service-type",
        "service_types",
        multiple=True,
        help="mDNS service type to query (repeatable, default: _http._tcp, _psia._tcp, _CGI._tcp)",
    )(func)
    return func



import json
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional

import numpy as np
import ultralytics
import yaml
from ultralytics import cfg as ultralytics_cfg

from highlighter.client import TrainingConfigType, json_tools
from highlighter.client.evaluation import (
    EvaluationMetric,
    EvaluationMetricCodeEnum,
    EvaluationMetricResult,
    find_or_create_evaluation_metric,
)
from highlighter.client.gql_client import HLClient
from highlighter.client.io import multithread_graphql_file_download
from highlighter.datasets.cropping import CropArgs
from highlighter.datasets.formats.yolo.writer import YoloWriter
from highlighter.trainers.base_trainer import BaseTrainer

__all__ = ["YoloV11Trainer"]


class YoloV11Trainer(BaseTrainer):

    def __init__(
        self,
        training_run_dir: Path,
        highlighter_training_config: TrainingConfigType,
    ):
        super().__init__(training_run_dir, highlighter_training_config)
        self._cfg = self._get_config()

    def _combine_hl_datasets(self, datasets):
        combined_ds = super()._combine_hl_datasets(datasets)

        # Ultralytics name their dataset splits differently so
        # we need to map them
        #   their "val" is our "test"
        #   their "test" is our "dev"
        combined_ds.data_files_df.loc[combined_ds.data_files_df.split == "test", "split"] = "val"
        combined_ds.data_files_df.loc[combined_ds.data_files_df.split == "dev", "split"] = "test"
        return combined_ds

    def _get_hl_metrics(self, client: HLClient):

        task = self._cfg["task"]

        if self._cfg["task"] in ("detect", "segment"):
            _metrics = self._get_metrics_det_seg()
        elif task == "classify":
            _metrics = self._get_metrics_classify()
        else:
            raise SystemExit(f"Invalid yolo task '{task}'")

        metrics = find_or_create_evaluation_metric(
            client,
            _metrics,
        )
        return {m.name: m for m in metrics}

    def _get_metrics_det_seg(self):
        cats = [c[1] for c in self.get_categories()]
        _metrics = []
        _metrics.extend(
            [
                EvaluationMetric(
                    research_plan_id=self.research_plan_id,
                    code=EvaluationMetricCodeEnum.mAP,
                    chart="Per Class Metrics",
                    description=f"Mean Avarage Precision ({cat.short_str()})",
                    iou=0.5,
                    name=f"mAP@IOU50({cat.short_str()})",
                    object_class_uuid=cat,
                )
                for cat in cats
            ]
        )
        _metrics.extend(
            [
                EvaluationMetric(
                    research_plan_id=self.research_plan_id,
                    code=EvaluationMetricCodeEnum.Other,
                    chart="Per Class Metrics",
                    description=f"Precision ({cat.short_str()})",
                    name=f"Precision({cat.short_str()})",
                    object_class_uuid=cat,
                )
                for cat in cats
            ]
        )
        _metrics.extend(
            [
                EvaluationMetric(
                    research_plan_id=self.research_plan_id,
                    code=EvaluationMetricCodeEnum.Other,
                    chart="Per Class Metrics",
                    description=f"Recall ({cat.short_str()})",
                    name=f"Recall({cat.short_str()})",
                    object_class_uuid=cat,
                )
                for cat in cats
            ]
        )

        _metrics.append(
            EvaluationMetric(
                research_plan_id=self.research_plan_id,
                code=EvaluationMetricCodeEnum.mAP,
                chart="Aggregate Metrics",
                description="Mean Avarage Precision Over All Classes",
                iou=0.5,
                name="mAP@IOU50",
            )
        )
        _metrics.append(
            EvaluationMetric(
                research_plan_id=self.research_plan_id,
                code=EvaluationMetricCodeEnum.Other,
                chart="Aggregate Metrics",
                description="Precision Over All Classes",
                name="Precision",
            )
        )
        _metrics.append(
            EvaluationMetric(
                research_plan_id=self.research_plan_id,
                code=EvaluationMetricCodeEnum.Other,
                chart="Aggregate Metrics",
                description="Recall Over All Classes",
                name="Recall",
            )
        )
        _metrics.append(
            EvaluationMetric(
                research_plan_id=self.research_plan_id,
                code=EvaluationMetricCodeEnum.Other,
                chart="Model Size",
                description="Size of the model, 0:Nano|1:Small|2:Medium|3:Large|4:XLarge",
                name="Model Size",
            )
        )
        _metrics.append(
            EvaluationMetric(
                research_plan_id=self.research_plan_id,
                code=EvaluationMetricCodeEnum.Other,
                chart="Ideal Confidence Threshold",
                description="The Confidence Threshold that maximizes the F1Score",
                name="Ideal Confidence Threshold",
            )
        )
        return _metrics

    def _get_metrics_classify(self):
        _metrics = []
        _metrics.append(
            EvaluationMetric(
                research_plan_id=self.research_plan_id,
                code=EvaluationMetricCodeEnum.Other,
                chart="Aggregate Metrics",
                description="Top One Accuracy",
                name="Accuracy Top1",
            )
        )
        _metrics.append(
            EvaluationMetric(
                research_plan_id=self.research_plan_id,
                code=EvaluationMetricCodeEnum.Other,
                chart="Aggregate Metrics",
                description="Top Five Accuracy",
                name="Accuracy Top5",
            )
        )
        _metrics.append(
            EvaluationMetric(
                research_plan_id=self.research_plan_id,
                code=EvaluationMetricCodeEnum.Other,
                chart="Model Size",
                description="Size of the model, 0:Nano|1:Small|2:Medium|3:Large|4:XLarge",
                name="Model Size",
            )
        )
        return _metrics

    def _get_config(self) -> Dict:
        if (self._training_run_dir / "cfg.yaml").exists():
            cfg = ultralytics.utils.YAML.load(self._training_run_dir / "cfg.yaml")
        else:
            cfg = self._get_default_config()
        return cfg

    def _get_default_config(self):
        overrides_lookup = {
            TrainingConfigType.TrainerType.YOLO_DET: {
                "model": "yolov8m.pt",
                "task": "detect",
            },
            TrainingConfigType.TrainerType.YOLO_SEG: {
                "model": "yolov8m-seg.pt",
                # Ensures overlapping masks are merged, which is critical for accurate
                # segmentation in datasets with overlapping objects (default is True, but
                # explicitly set for clarity).
                "overlap_mask": True,
                # Controls mask resolution; 4 is the default and balances detail
                # with computational efficiency, suitable for most segmentation tasks.
                "mask_ratio": 4,
                "task": "segment",
            },
            TrainingConfigType.TrainerType.YOLO_CLS: {
                "model": "yolov8m-cls.pt",
                # Adds regularization to prevent overfitting, which is more common
                # in classification tasks with large datasets; 0.1 is a modest starting point.
                "dropout": 0.1,
                # Classification often works with smaller images than detection/segmentation;
                # 224 is a common size (e.g., ImageNet), balancing detail and speed.
                "imgsz": 224,
                "task": "classify",
            },
        }

        overrides = overrides_lookup[self._hl_training_config.trainer_type]

        default_cfg = dict(ultralytics_cfg.get_cfg())
        default_cfg.update(overrides)
        default_cfg["project"] = "runs"
        default_cfg["opset"] = 14
        default_cfg["format"] = "onnx"
        default_cfg["dynamic"] = False
        return default_cfg

    def generate_boilerplate(self):
        super().generate_boilerplate()
        with (self._training_run_dir / "cfg.yaml").open("w") as f:
            yaml.dump(self._cfg, f)

    @property
    def config_path(self):
        return self._training_run_dir / "cfg.yaml"

    def _make_classify_artefact(self, onnx_filepath) -> Path:
        crop_args = self.get_crop_args()
        if isinstance(crop_args, CropArgs):
            crop_args_dict = crop_args.model_dump()
        elif crop_args is None:
            crop_args_dict = None
        else:
            raise ValueError(f"Crop args must be None or an instance of `CropArgs` got: {crop_args}")
        d = dict(
            file_url=str(Path(onnx_filepath.absolute())),
            type="OnnxOpset14",
            inference_config=dict(
                type="classifier",
                code="BoxClassifier",
                machine_agent_type_id="d4787671-3839-4af9-9b34-a686faafbfae",
                parameters=dict(
                    output_format="yolov8_cls",
                    cropper=crop_args_dict,
                ),
            ),
            training_config=self._cfg,
        )

        artefact_path = onnx_filepath.parent / "artefact.yaml"
        with artefact_path.open("w") as f:
            yaml.dump(d, f)

        return artefact_path

    def _make_detect_segment_artefact(self, onnx_filepath) -> Path:
        output_format = "yolov8_seg" if self._cfg["task"] == "segment" else "yolov8_det"
        d = dict(
            file_url=str(Path(onnx_filepath.absolute())),
            type="OnnxOpset14",
            inference_config=dict(
                type="detector",
                code="Detector",
                machine_agent_type_id="29653174-8f45-440d-b75a-4ed0aa5fa6ff",
                parameters=dict(
                    output_format=output_format,
                ),
            ),
            training_config=self._cfg,
        )

        artefact_path = onnx_filepath.parent / "artefact.yaml"
        with artefact_path.open("w") as f:
            yaml.dump(d, f)

        return artefact_path

    def export_to_onnx(self, trained_model) -> Path:
        return Path(trained_model.export(format="onnx", batch=1, dynamic=False, device=0))

    def make_artefact(self, onnx_file_path: Path) -> Path:
        # Disable ultralyitcs' auto install of packages
        ultralytics.utils.checks.AUTOINSTALL = False

        _task = self._cfg["task"]
        if _task == "classify":
            return self._make_classify_artefact(onnx_file_path)
        elif _task in ("detect", "segment"):
            return self._make_detect_segment_artefact(onnx_file_path)
        else:
            raise ValueError(f"Invalid yolo task '{_task}', expected one of (classify|detect|segment)")

    def train(self) -> Any:

        model = ultralytics.YOLO(self._cfg["model"])

        ultralytics.settings.update({"datasets_dir": str(self._training_run_dir.absolute())})

        data_cfg_path = (self._training_run_dir / "datasets" / "data.yaml").absolute()
        with data_cfg_path.open("r") as f:
            data_cfg = yaml.safe_load(f)

        self._cfg["data"] = str(data_cfg_path)
        self._cfg["single_cls"] = data_cfg["nc"] == 1

        if self._cfg["task"] == "classify":
            self._cfg["data"] = str(data_cfg_path.parent)
            self._cfg["classes"] = None  # Classification classes are specified via directories
        else:
            self._cfg["classes"] = list(data_cfg["names"].keys())
            self._cfg["data"] = str(data_cfg_path)

        trainer = None

        if getattr(self._hl_training_config, "oversample", True):
            task = self._cfg.get("task")

            if task == "classify":
                from highlighter.trainers.yolov11._weighted_sampling import (
                    WeightedClassificationTrainer,
                )

                trainer = WeightedClassificationTrainer

            elif task == "detect" or task == "segment":
                from highlighter.trainers.yolov11._weighted_sampling import (
                    WeightedDetectionTrainer,
                )

                trainer = WeightedDetectionTrainer

            else:
                raise ValueError(
                    f"Oversampling is enabled but task '{task}' is not supported. "
                    "Supported tasks: ['classify', 'detect', 'segment']"
                )

        model.train(trainer=trainer, **self._cfg)
        return model

    def _export(self, checkpoint, cfg_overrides={}):

        self._cfg["model"] = checkpoint
        self._cfg.update(cfg_overrides)
        model = Path(self._cfg["model"])
        artefact_path = self.make_artefact(model)

        return artefact_path.absolute()

    def get_crop_args(self) -> Optional[CropArgs]:
        if self._cfg["task"] == "classify":
            crop_args = self._hl_training_config.crop_args
        else:
            crop_args = None
        return crop_args

    @property
    def training_data_dir(self) -> Path:
        return self._training_run_dir / "datasets"

    def generate_trainer_specific_dataset(self, hl_dataset):

        if not (self.training_data_dir / "data.yaml").exists():
            self.training_data_dir.mkdir(exist_ok=True)
            # Optionally filter dataset, see filter_dataset's doc str
            filtered_hl_ds = self.filter_dataset(hl_dataset)

            # Download required images
            image_cache_dir = self._training_run_dir / "images"
            multithread_graphql_file_download(
                HLClient.get_client(),
                filtered_hl_ds.data_files_df.data_file_id.values,
                image_cache_dir,
            )

            ddf = filtered_hl_ds.data_files_df
            if any([Path(f).suffix.lower() == ".mp4" for f in ddf.filename.unique()]):

                print("Detected video dataset, interpolating data from keyframes")
                filtered_hl_ds = filtered_hl_ds.interpolate_from_key_frames(
                    frame_save_dir=image_cache_dir,
                    source_file_dir=image_cache_dir,
                )

            # Write dataset in yolo format
            writer = YoloWriter(
                output_dir=self.training_data_dir,
                image_cache_dir=image_cache_dir,
                categories=self.get_categories(),
                task=self._cfg["task"],
                crop_args=self.get_crop_args(),
            )
            writer.write(filtered_hl_ds)

    def evaluate(
        self, checkpoint: Path | str, cfg_path: Optional[Path | str] = None
    ) -> Dict[str, EvaluationMetricResult]:
        """Evaluate a YOLOv11 model and generate structured evaluation metrics.

        Evaluates the model using ultralytics validation and creates evaluation metric
        result objects for upload to Highlighter. Handles both classification and
        detection/segmentation tasks with appropriate metrics for each.

        Args:
            checkpoint: Path to model checkpoint (can be .pt, .onnx, or other supported formats)
            cfg_path: Optional path to configuration file (uses self._cfg if None)

        Returns:
            List[EvaluationMetricResult]: Evaluation metric results ready for upload

        Side Effects:
            Saves evaluation results to 'eval_metric_results.json' in checkpoint directory
        """
        eval_metric_results: Dict[str, EvaluationMetricResult] = {}
        if cfg_path is None:
            cfg: Dict = self._cfg
        else:
            cfg: Dict = ultralytics.utils.YAML.load(str(cfg_path))

        assert cfg["data"] is not None
        client = HLClient.get_client()

        model = ultralytics.YOLO(str(checkpoint), task=cfg["task"])
        results = model.val(**cfg)

        if cfg["task"] in ("segment", "detect"):
            eval_metric_results = self._create_eval_results_det_or_seg(results, cfg, client)
        elif cfg["task"] == "classify":
            eval_metric_results = self._create_eval_results_cls(results, cfg, client)

        with (Path(checkpoint).parent / "eval_metric_results.json").open("w") as f:
            json.dump(
                {k: e.model_dump() for k, e in eval_metric_results.items()},
                f,
                indent=2,
                cls=json_tools.HLJSONEncoder,
            )

        return eval_metric_results

    def _create_eval_results_cls(
        self, results, cfg: Dict, client: HLClient
    ) -> Dict[str, EvaluationMetricResult]:
        """Create evaluation metric results for classification tasks.

        Extracts classification-specific metrics including model size, top-1 accuracy,
        and top-5 accuracy from ultralytics validation results.

        Args:
            results: Ultralytics validation results object
            cfg: Configuration dictionary containing model information
            client: HLClient for metric lookup

        Returns:
            List[EvaluationMetricResult]: Classification evaluation metrics
        """
        eval_metric_results: Dict[str, EvaluationMetricResult] = {}
        _hl_metrics = self._get_hl_metrics(client)

        model_size_char = Path(cfg["model"]).stem.replace("-cls", "")[-1]
        if model_size_char in "nsmlx":
            model_size_int = "nsmlx".index(model_size_char)
            eval_metric_results["Model Size"] = _hl_metrics["Model Size"].result(
                model_size_int, self.training_run_id
            )
        else:
            model_str = cfg["model"]
            raise SystemExit(f"Unvalid model_size_char '{model_size_char}' from '{model_str}'")
        acc_top1 = results.results_dict["metrics/accuracy_top1"]
        acc_top5 = results.results_dict["metrics/accuracy_top5"]
        eval_metric_results["Accuracy Top1"] = _hl_metrics["Accuracy Top1"].result(
            acc_top1, self.training_run_id
        )
        eval_metric_results["Accuracy Top5"] = _hl_metrics["Accuracy Top5"].result(
            acc_top5, self.training_run_id
        )
        return eval_metric_results

    def _create_eval_results_det_or_seg(
        self, results, cfg: Dict, client: HLClient
    ) -> Dict[str, EvaluationMetricResult]:
        """Create evaluation metric results for detection and segmentation tasks.

        Extracts detection/segmentation-specific metrics including precision, recall,
        mAP@IoU50, model size, ideal confidence threshold, and per-class mAP values
        from ultralytics validation results.

        Args:
            results: Ultralytics validation results object
            cfg: Configuration dictionary containing model information
            client: HLClient for metric lookup

        Returns:
            List[EvaluationMetricResult]: Detection/segmentation evaluation metrics

        Note:
            Uses corrected ultralytics result attributes (e.g., 'metrics/precision(B)')
            rather than deprecated mean_results() method for better accuracy.
        """
        eval_metric_results: Dict[str, EvaluationMetricResult] = {}
        _hl_metrics = self._get_hl_metrics(client)

        eval_metric_results["Precision"] = _hl_metrics["Precision"].result(
            results.results_dict["metrics/precision(B)"], self.training_run_id
        )
        eval_metric_results["Recall"] = _hl_metrics["Recall"].result(
            results.results_dict["metrics/recall(B)"], self.training_run_id
        )
        eval_metric_results["mAP@IOU50"] = _hl_metrics["mAP@IOU50"].result(
            results.results_dict["metrics/mAP50(B)"], self.training_run_id
        )

        model_size_char = Path(cfg["model"]).stem.replace("-det", "").replace("-seg", "")[-1]
        if model_size_char in "nsmlx":
            model_size_int = "nsmlx".index(model_size_char)
            eval_metric_results["Model Size"] = _hl_metrics["Model Size"].result(
                model_size_int, self.training_run_id
            )
        else:
            model_str = cfg["model"]
            raise SystemExit(f"Unvalid model_size_char '{model_size_char}' from '{model_str}'")

        f1_conf_curve_idx = results.curves.index("F1-Confidence(B)")

        if len(results.curves_results[f1_conf_curve_idx][1].shape) == 2:
            best_f1_idx = np.argmax(results.curves_results[f1_conf_curve_idx][1].mean(axis=0))
        else:
            best_f1_idx = np.argmax(results.curves_results[f1_conf_curve_idx][1])
        best_f1_thr = results.curves_results[f1_conf_curve_idx][0][best_f1_idx]
        eval_metric_results["Ideal Confidence Threshold"] = _hl_metrics["Ideal Confidence Threshold"].result(
            best_f1_thr, self.training_run_id
        )

        cats = [c[1] for c in self.get_categories()]
        CLS_PRECISION_IOU50 = 2
        for cls_idx, cat in enumerate(cats):
            class_name = cat.short_str()
            metric_name = f"mAP@IOU50({class_name})"

            try:
                value = results.class_result(cls_idx)[CLS_PRECISION_IOU50]
            except IndexError as _:
                value = 0.0

            eval_metric_results[metric_name] = _hl_metrics[metric_name].result(value, self.training_run_id)

        return eval_metric_results



# Usage
# ~~~~~
# grep -A14 graph example.json  # Show PipelineDefinition Graph
# grep '{ "name"' example.json  # Show PipelineElements list
#
# To Do
# ~~~~~
# - Improve "Mock" --> "Code" and/or "LISP" :)
#   - Parameter "code": "{output} = {input} + {increment}"

from typing import Tuple

import aiko_services as aiko

__all__ = ["Mock", "NoOp"]

# --------------------------------------------------------------------------- #

class Mock(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("mock:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream) -> Tuple[aiko.StreamEvent, dict]:
        label, _ = self.get_parameter("label", default="")
        self.logger.debug(f"{self.my_id()} {label}")
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #

class NoOp(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("noop:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream) -> Tuple[aiko.StreamEvent, dict]:
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #



class OptionalPackageMissingError(Exception):
    def __init__(self, package_name, option_name):
        super().__init__(f"{package_name} not found. Use pip install highlighter-sdk[{option_name}]")


def require_package(package, package_name, option_name):
    """
    Decorator that checks if a specific module is available before running a function.

    :param module: The module to check (should be None if not imported).
    :param module_name: The name of the module as a string.
    :raises OptionalImportException: If the module is not available.
    """
    packages = package if isinstance(package, (list, tuple)) else [package]
    package_names = package_name if isinstance(package_name, (list, tuple)) else [package_name]
    option_names = option_name if isinstance(option_name, (list, tuple)) else [option_name]

    def decorator(func):
        def wrapper(*args, **kwargs):
            for m, pn, on in zip(packages, package_names, option_names):
                if m is None:
                    raise OptionalPackageMissingError(pn, on)
            return func(*args, **kwargs)

        return wrapper

    return decorator



# Definition(s)
# ~~~~~~~~~~~~~
# swag: One's personal bundle of belongings (Australian lingo)
# SWAG: Stuff We All Get
# SWAG: https://en.wikipedia.org/wiki/Scientific_wild-ass_guess :)
#
# Notes
# ~~~~~
# - Watch out for "stream.py:Frame" dataclass and
#   "dashboard.py" use of "asciimatics.widgets.Frame()" class
#
# - If required, Stream.remote_graph_path() used to determine remote Graph Path
#
# To Do
# ~~~~~
# - Implement dataclass "Class Metrics" to replace "metrics" dictionary
#
# - Refactor from "pipeline.py", extract Stream concepts including Parameters
#   - Review "../archive/main/stream_2020.py"

from dataclasses import dataclass, field
import queue
from typing import Any, Dict

from aiko_services.main.utilities import *

__all__ = [
    "DEFAULT_STREAM_ID", "FIRST_FRAME_ID", "Frame", "Stream",
    "StreamEvent", "StreamEventName", "StreamState", "StreamStateName"
]

DEFAULT_STREAM_ID = "*"  # string
FIRST_FRAME_ID = 0       # integer

class StreamEvent:
    ERROR      =   -2  # Move to StreamState.ERROR
    STOP       =   -1  # Move to StreamState.STOP
    OKAY       =    0  # Stay calm and keep on running
    NO_FRAME   =    1  # No frame to process and keep on running
    DROP_FRAME =    2  # No longer process this frame and keep on running
    LOOP_END   =    3  # Complete ControlFlowLoop and keep on running
    USER       = 1024  # User defined custom StreamEvents start from here

StreamEventName = {
    StreamEvent.DROP_FRAME: "DropFrame",
    StreamEvent.ERROR:      "Error",
    StreamEvent.LOOP_END:   "LoopEnd",
    StreamEvent.NO_FRAME:   "NoFrame",
    StreamEvent.OKAY:       "Okay",
    StreamEvent.STOP:       "Stop",
    StreamEvent.USER:       "User"
}

class StreamState:
    ERROR       =   -2  # Don't generate new frames and ignore queued frames
    STOP        =   -1  # Don't generate new frames and process queued frames
    RUN         =    0  # Generate new frames and process queued frames
    NO_FRAME    =    1  # No frame to process, then back to RUN state
    DROP_FRAME  =    2  # Stop processing current frame, then back to RUN state
    USER        = 1024  # User defined custom StreamStates start from here

StreamStateName = {
    StreamState.DROP_FRAME: "DropFrame",
    StreamState.ERROR:      "Error",
    StreamState.NO_FRAME:   "NoFrame",
    StreamState.STOP:       "Stop",
    StreamState.RUN:        "Run",
    StreamState.USER:       "User"
}

@dataclass
class Frame:  # effectively a continuation :)
    metrics: Dict[str, Any] = field(default_factory=dict)  # TODO: Dataclass
    paused_pe_name: str = None  # remote PipelineElement that has been called
### self.pipeline_graph.iterate_after(frame.paused_pe_name, frame.graph_path)
### graph_path: str = None  # Graph path (head_node_name)
    swag: Dict[str, Any] = field(default_factory=dict)  # PipelineElement data

@dataclass
class Stream:
    stream_id: str = DEFAULT_STREAM_ID
    frame_id: int = FIRST_FRAME_ID  # only updated by Pipeline thread
    frames: Dict[int, Frame] = field(default_factory=dict)
    graph_path: str = None  # Graph path (head_node_name), default: first path
    lock: Lock = field(init=False)
    parameters: Dict[str, Any] = field(default_factory=dict)
    queue_response: queue = None
    state: StreamState = StreamState.RUN
    topic_response: str = None
    variables: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        self.lock = Lock(f"{__name__}_{self.stream_id}")

    def set_state(self, state):
        if state == StreamState.ERROR and self.state > StreamState.ERROR:
            self.state = state
        if state == StreamState.STOP and self.state > StreamState.STOP:
            self.state = state
        else:
            self.state = state

# https://docs.python.org/3/library/dataclasses.html#dataclasses.asdict
    def as_dict(self):
    #   return {key: str(value) for key, value in asdict(self).items()}
        return {
            "stream_id": self.stream_id,
            "frame_id": self.frame_id,
            "graph_path": self.graph_path
        }

# https://docs.python.org/3/library/dataclasses.html#dataclasses.replace
    def update(self, stream_dict):
        if not isinstance(stream_dict, dict):
            return False
    #   self = replace(self, **stream_dict)
        self.stream_id = str(stream_dict.get("stream_id", self.stream_id))
        self.frame_id = int(stream_dict.get("frame_id", self.frame_id))
        self.graph_path = stream_dict.get("graph_path", self.graph_path)
        self.parameters = stream_dict.get("parameters", self.parameters)
        self.queue_response = stream_dict.get(
            "queue_response", self.queue_response)
        self.state = int(stream_dict.get("state", StreamState.RUN))
        self.topic_response = stream_dict.get(
            "topic_response", self.topic_response)
        return True



# To Do
# ~~~~~
# * Rename TransportMQTT --> MQTTActor
#   Rename "transport_mqtt.py" --> "mqtt_actor.py"
#
# - Design and implementation for Ray, ROS2 and more

from inspect import getmembers, isfunction

from aiko_services.main import *
from aiko_services.main.utilities.parser import generate

__all__ = [
    "TransportMQTT", "TransportMQTTImpl","ActorDiscovery", "get_actor_mqtt"
]

_LOGGER = aiko.logger(__name__)

class TransportMQTT(Actor):
    Interface.default("TransportMQTT",
        "aiko_services.main.transport.transport_mqtt.TransportMQTTImpl")

class TransportMQTTImpl(TransportMQTT):
    def __init__(self, context):
        context.call_init(self, "Actor", context)

    def terminate(self):
        self.stop()

class ServiceDiscovery:  # Move to registrar.py or share.py ?
    pass                 # Refactor after ActorDiscovery starts to work properly

class ActorDiscovery(ServiceDiscovery):  # Move to actor.py or share.py ?
    def __init__(self, service):
        self.services_cache = services_cache_create_singleton(service)

    def add_handler(self, service_change_handler, filter):
        self.services_cache.add_handler(service_change_handler, filter)

    def remove_handler(self, service_change_handler, filter):
        self.services_cache.remove_handler(service_change_handler, filter)

# TODO: Update to use ServiceFields.name, rather than "actor=name" tag
    def get_actor_mqtt(self, name):
        actor_topic = ".".join(name.split(".")[:-1])  # WIP: Actor name
        services = self.services_cache.get_services()
        raise Exception("Broken: get_actor_mqtt()")  # REVIEW FIXME
        services = services.filter_by_name(name)
        actor = services.get(name)
        return actor

# -----------------------------------------------------------------------------

def create_actor_mqtt(
    actor_class, name, actor_init_args={}, resources=None, daemon = True):
    pass

def delete_actor_mqtt(actor):
    actor.terminate()

def get_public_methods(protocol_class):
    if isinstance(protocol_class, str):
        raise ValueError(
            f"{protocol_class} is a String, should be a Class reference ?")
    public_method_names = [
        method_name
        for method_name, method in getmembers(protocol_class, isfunction)
        if not method_name.startswith("_")
    ]
    if len(public_method_names) == 0:
        raise ValueError(f"Class {protocol_class} has no public methods")
    return public_method_names

def make_proxy_mqtt(target_topic_in, public_method_names):
    class ServiceRemoteProxy(): pass

    def _proxy_send_message(method_name):
        def closure(*args, **kwargs):
            parameters = args if not kwargs else [args[0], kwargs]
            payload = generate(method_name, parameters)
            aiko.message.publish(target_topic_in, payload)
        return closure

    service_remote_proxy = ServiceRemoteProxy()
    for method_name in public_method_names:
        setattr(service_remote_proxy,
            method_name, _proxy_send_message(method_name))
    return service_remote_proxy

def get_actor_mqtt(target_service_topic_in, protocol_class):
    public_methods = get_public_methods(protocol_class)
    actor_proxy = make_proxy_mqtt(target_service_topic_in, public_methods)
    return actor_proxy

# -----------------------------------------------------------------------------



# Usage: File
# ~~~~~~~~~~~
# aiko_pipeline create pipelines/text_pipeline_0.json -s 1 -sr -ll debug
#
# aiko_pipeline create pipelines/text_pipeline_0.json -s 1 -p rate 1.0
#
# aiko_pipeline create pipelines/text_pipeline_0.json -s 1  \
#   -p TextReadFile.data_batch_size 8
#
# aiko_pipeline create pipelines/text_pipeline_0.json -s 1  \
#   -p TextReadFile.data_sources file://data_in/in_{}.txt
#
# aiko_pipeline create pipelines/text_pipeline_0.json -s 1  \
#     -p TextWriteFile.path "file://data_out/out_{:02d}.txt"
#
# aiko_pipeline create pipelines/text_pipeline_0.json -s 1 \
#   -p TextReadFile.data_sources file://data_in/in_00.txt  \
#   -p TextTransform.transform titlecase                   \
#   -p TextWriteFile.data_targets file://data_out/out_00.txt
#
# Usage: Drop frame tests (local and remote)
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# aiko_pipeline create pipelines/text_pipeline_1.json -s 1 -ll debug
#
# aiko_pipeline create pipelines/text_pipeline_2.json -s 1 -ll debug  # local
# aiko_pipeline create pipelines/text_pipeline_3.json      -ll debug  # remote
#
# Usage: ZMQ
# ~~~~~~~~~~
# aiko_pipeline create pipelines/text_zmq_pipeline_0.json -s 1 -sr  \
#            -ll debug -gt 10
#
# aiko_pipeline create pipelines/text_zmq_pipeline_0.json -s 1 -sr  \
#            -p TextReadZMQ.data_sources zmq://0.0.0.0:6502
#
# aiko_pipeline create pipelines/text_zmq_pipeline_1.json -s 1 -sr  \
#            -ll debug                                              \
#            -p TextReadFile.rate 2.0                               \
#            -p TextWriteZMQ.data_targets zmq://192.168.0.1:6502
#
# Resources
# ~~~~~~~~~
# - https://learning-0mq-with-pyzmq.readthedocs.io/en/latest/pyzmq/patterns/patterns.html
#
# To Do
# ~~~~~
# * ZMQ remote PipeElement that sends "content" out-of-band via ZMQ
#   * Compared to the default remote PipelineElement that uses in-band via MQTT
#   * Effectively ZMQ as another transport implementation :)
#
# - Support for "media type" encoding details for "text"
#   - Consider additional encoding information in out-of-band text records ?
#     - "frame_id" and/or "text:length:content" or "text/zip:length:content" ?
#
# - TextReadFile(s): Single file or list of files or directory
#   - Option: Each line is a record (streaming)
#   - Formatted as CR/LF records, JSON, XML, CSV
#
# - TextWriteFile(s): Single file or list of files or directory
#   - Option: Each line is a record (streaming)
#   - Formatted as CR/LF records, JSON, XML, CSV
#
# - TextFilter: line/word/character count, ...
# - TextTransform: strip(), ...
#
# - Rather than TextSample by frame ... sample by text count in [texts] input
#
# - Pre-processing / Post-processing, e.g abbrevations, acronyms
#   - Speech-To-Text: Words not easily recognised
#     - Example: Microphone --> Speech-To-Text --> Text
#   - Text-To-Speech: Poor pronouciation
#     - Example: Text --> Text-To-Speech --> Speaker
#
# - Text Framing for LLM

from typing import Tuple
from pathlib import Path

import aiko_services as aiko

__all__ = [
    "TextOutput", "TextReadFile", "TextReadZMQ",
    "TextSample", "TextTransform", "TextWriteFile", "TextWriteZMQ"
]

# --------------------------------------------------------------------------- #
# Useful for Pipeline output that should be all of the text processed

class TextOutput(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("text_output:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, texts) -> Tuple[aiko.StreamEvent, dict]:
        return aiko.StreamEvent.OKAY, {"texts": texts}

# --------------------------------------------------------------------------- #
# TextReadFile is a DataSource which supports ...
# - Individual text files
# - Directory of text files with an optional filename filter
# - TODO: Archive (tgz, zip) of text files with an optional filename filter
#
# parameter: "data_sources" is the read file path, format variable: "frame_id"
#
# Note: Only supports Streams with "data_sources" parameter

class TextReadFile(aiko.DataSource):  # PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("text_read_file:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, paths) -> Tuple[aiko.StreamEvent, dict]:
        texts = []
        for path in paths:
            try:
                with path.open("r") as file:
                    text = file.read()
                texts.append(text)
                self.logger.debug(f"{self.my_id()}: {path} ({len(text)})")
            except Exception as exception:
                diagnostic = f"Error loading text: {exception}"
                return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        return aiko.StreamEvent.OKAY, {"texts": texts}

# --------------------------------------------------------------------------- #

class TextReadTTY(aiko.DataSource):  # PipelineElement
    def __init__(self, context):
        context.set_protocol("text_read_tty:0")
        context.call_init(self, "PipelineElement", context)

    def start_stream(self, stream, stream_id):
        print('# Type "/?" for help and "/x" to exit')
        self.commands = {
            "/?":   (self._command_help,    "Help"),
            "//":   (None,                  "Escape sequence for '/'"),
            "/h":   (self._command_history, "List command line history"),
            "/h N": (self._command_history, "Execute command line 'N'"),
            "/x":   (self._command_exit,    "Exit")
        }
        stream.variables["tty_command_lines"] = []
        return super().start_stream(stream, stream_id)

    def _command_exit(self, stream, tokens):
        raise SystemExit("Process terminated")

    def _command_help(self, stream, tokens):
        for command, command_details in self.commands.items():
            print(f"{command:4s}: {command_details[1]}")
        return None

    def _command_history(self, stream, tokens):
        command_line = None
        tty_command_lines = stream.variables["tty_command_lines"]
        if len(tokens):
            if tokens[0].isdigit():
                index = int(tokens[0])
                if index < len(tty_command_lines):
                    command_line = tty_command_lines[index]
        else:
            for index, tty_command_line in enumerate(tty_command_lines):
                print(f"{index:03d}: {tty_command_line}")
        return command_line

    def _parse_command_line(self, stream, text):
        tokens = text.split()
        command = tokens[0] if len(tokens) else ""
        command_line = None
        if command in self.commands:
            command_function = self.commands[command][0]
            if command_function:
                command_line = command_function(stream, tokens[1:])
            else:
                command_line = text[3:]
        return command_line

    def process_frame(self, stream, records) -> Tuple[aiko.StreamEvent, dict]:
        texts = []
        for text in records:
            command_line = text
            save = True
            if command_line.startswith("/"):
                command_line = self._parse_command_line(stream, command_line)
                save = False
            if command_line:
                if save:
                    stream.variables["tty_command_lines"].append(command_line)
                texts.append(command_line)
        return aiko.StreamEvent.OKAY, {"texts": texts}
    #   return aiko.StreamEvent.DROP_FRAME, {}  # TODO: If command_line is None

# --------------------------------------------------------------------------- #
# TextReadZMQ is a DataSource which supports ...
# - TextWriteZMQ(DataTarget) ZMQ client --> TextReadZMQ(DataSource) ZMQ server
#   - Individual text records produced by ZMQ client and consumed by ZMQ server
#
# parameter: "data_sources" is the ZMQ server bind details (scheme_zmq.py)
#
# Note: Only supports Streams with "data_sources" parameter

class TextReadZMQ(aiko.DataSource):  # PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("text_read_zmq:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, records) -> Tuple[aiko.StreamEvent, dict]:
        texts = []
        for record in records:
            text = record.decode()
    #       if text.startswith("text:"):  # TODO: "text:length:content" ?
    #           tokens = text.split(":")
    #           text = tokens[2:][0]      # just the "content"
            texts.append(text)
            self.logger.debug(f"{self.my_id()}: {text} ({len(text)})")
        return aiko.StreamEvent.OKAY, {"texts": texts}

# --------------------------------------------------------------------------- #

class TextSample(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("text_sample:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, texts) -> Tuple[aiko.StreamEvent, dict]:
        sample_rate, _ = self.get_parameter("sample_rate", 1)
        if stream.frame_id % sample_rate:
            self.logger.debug(f"{self.my_id()}: frame dropped")
            return aiko.StreamEvent.DROP_FRAME, {}
        else:
            self.logger.debug(f"{self.my_id()}: frame not dropped")
            return aiko.StreamEvent.OKAY, {"texts": texts}

# --------------------------------------------------------------------------- #

class TextTransform(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("text_transform:0")
        context.call_init(self, "PipelineElement", context)

        self.transforms = {
            "lowercase": lambda text: text.lower(),  # looks like this !
            "none":      lambda text: text,          # looks unchanged !
            "titlecase": lambda text: text.title(),  # Looks Like This !
            "uppercase": lambda text: text.upper()   # LOOKS LIKE THIS !
        }

    def process_frame(self, stream, texts) -> Tuple[aiko.StreamEvent, dict]:
        transform_type, found = self.get_parameter("transform")
        if not found:
            diagnostic = 'Must provide "transform" parameter'
            return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        transform = self.transforms.get(transform_type, None)
        if not transform:
            diagnostic = f"Unknown text transform type: {transform_type}"
            return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        texts_transformed = []
        if transform_type == "none":
            texts_transformed = texts  # optimization :)
        else:
            for text in texts:
                transformed_text = transform(text)
                texts_transformed.append(transformed_text)

        return aiko.StreamEvent.OKAY, {"texts": texts_transformed}

# --------------------------------------------------------------------------- #
# TextWriteFile is a DataTarget with a text string parameter
#
# parameter: "data_targets" is the write file path, format variable: "frame_id"
#
# stream.variables["target_path_template"] indicates whether ...
# - False: All text records should be written to the same "target_path"
# - True:  Each text record is written to a different "target_path_template"
#
# Note: Only supports Streams with "data_targets" parameter

class TextWriteFile(aiko.DataTarget):  # PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("text_write_file:0")
        context.call_init(self, "PipelineElement", context)

    def start_stream(self, stream, stream_id):
        stream_event, diagnostic = super().start_stream(stream, stream_id)
        if stream_event != aiko.StreamEvent.OKAY:
            return stream_event, diagnostic

        if not stream.variables["target_path_template"]:
            path = stream.variables["target_path"]
            self.logger.debug(f"{self.my_id()}: {path}")
            try:
                stream.variables["target_file"] = Path(path).open("w")
            except Exception as exception:
                diagnostic = f"Error saving text: {exception}"
                return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        return aiko.StreamEvent.OKAY, {}

    def process_frame(self, stream, texts) -> Tuple[aiko.StreamEvent, dict]:
        for text in texts:
            if stream.variables["target_path_template"]:
                path = stream.variables["target_path"]
                path = path.format(stream.variables["target_file_id"])
                stream.variables["target_file_id"] += 1

                self.logger.debug(f"{self.my_id()}: {path}")
                try:
                    with Path(path).open("w") as file:
                        file.write(text)
                except Exception as exception:
                    diagnostic = f"Error saving text: {exception}"
                    return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}
            else:
                try:
                    stream.variables["target_file"].write(f"{text}")
                except Exception as exception:
                    diagnostic = f"Error saving text: {exception}"
                    return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        return aiko.StreamEvent.OKAY, {}

    def stop_stream(self, stream, stream_id):
        stream_event, diagnostic = super().stop_stream(stream, stream_id)
        if "target_file" in stream.variables:
            stream.variables["target_file"].close()
            del stream.variables["target_file"]
        return stream_event, diagnostic

# --------------------------------------------------------------------------- #

class TextWriteTTY(aiko.DataTarget):  # PipelineElement
    def __init__(self, context):
        context.set_protocol("text_write_tty:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, texts) -> Tuple[aiko.StreamEvent, dict]:
        for text in texts:
            print(text)
        command_lines_length = len(stream.variables["tty_command_lines"])
        tty_prompt, _ = self.get_parameter("tty_prompt", default="> ")
        tty_prompt = f"{command_lines_length:03d}:{tty_prompt}"
        print(f"{tty_prompt}", flush=True, end="")
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #
# TextWriteZMQ is a DataTarget which supports ...
# - TextWriteZMQ(DataTarget) ZMQ client --> TextReadZMQ(DataSource) ZMQ server
#   - Individual text records produced by ZMQ client and consumed by ZMQ server
#
# parameter: "data_targets" is the ZMQ connect details (scheme_zmq.py)
#
# Note: Only supports Streams with "data_targets" parameter

class TextWriteZMQ(aiko.DataTarget):  # PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("text_write_zmq:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, texts) -> Tuple[aiko.StreamEvent, dict]:
        media_type = "text"                            # TODO: "text/zip" ?
        for text in texts:
     #      text = f"{media_type}:{len(text)}:{text}"  # "text:length:content" ?
            stream.variables["target_zmq_socket"].send(text.encode())

        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #



from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from highlighter.core.exceptions import require_package

try:
    import cv2
except ModuleNotFoundError as _:
    cv2 = None
import numpy as np
from PIL import Image
from pydantic import BaseModel, field_validator
from shapely import affinity
from shapely import geometry as gm

PathLike = Union[str, Path]
LTRBRect = Tuple[int, int, int, int]

KEY_CROP_ROTATED_RECT = "crop_rotated_rect"

__all__ = [
    "CropArgs",
    "KEY_CROP_ROTATED_RECT",
    "crop_rect_from_poly",
    "crop_rectangle",
    "crop_rotated_rectangle",
]


class CropArgs(BaseModel):
    crop_rotated_rect: Optional[bool] = False
    scale: Optional[float] = None
    pad: Optional[int] = None
    warped_wh: Optional[Union[Tuple[int, int], List[int]]] = None

    @field_validator("warped_wh")
    @classmethod
    def _validate_warped_wh(cls, v):
        """If warped_wh is not None make sure it's a list and not a tuple
        because tuple is not json/safe_yaml serializable.
        """

        if isinstance(v, tuple):
            v = list(v)

        if v is not None:
            assert isinstance(v, list)
            assert len(v) == 2
        return v

    def dict(self, **kwargs):
        d = super().model_dump(**kwargs)
        if not self.crop_rotated_rect:
            d.pop("warped_wh")
        return d


def _get_wh_of_rotated_rect(np_corners: np.ndarray) -> Tuple[int, int]:
    # First 3 corners of the rectangle
    a, b, c = np_corners[0], np_corners[1], np_corners[2]
    ab = int(round(np.linalg.norm(b - a)))
    bc = int(round(np.linalg.norm(c - b)))
    if ab >= bc:
        w, h = ab, bc
    else:
        w, h = bc, ab

    return w, h


def _order_rect_points_bottom_left_ccw(rect: gm.Polygon) -> gm.Polygon:
    # 0. Note, the origin for Shapely is the bottom left with
    #    the positive x and y axis right and up respectivally.
    #    But, our coordinates have the origin at the top left
    #    and positive x,y being rigth and down.
    #    So our top right and clockwise will translate to shapely's
    #    Bottom left and counter-clockwise. See below for an illustration
    #
    #    Shapely coords:
    #
    #    +y
    #    |
    #    | 0,1      2,1
    #    * * * * * *
    #    *         *
    #    * * * * * * --- +x
    #     0,0      2,0

    #    Highlighter coords:
    #
    #     0,0      2,0
    #    * * * * * * --- +x
    #    *         *
    #    * * * * * *
    #    |0,1      2,1
    #    |
    #    +y
    #
    #
    # The following uses the shapely coordinate system, that way, if we
    # look for bottom left and order ccw we should be golden
    #
    # 1. Find the edges that belong to the major axis (longest)
    # 2. Select the bottom edge
    # 3. Find the left most point of the bottom edge
    # 4. Set bottom left point to 0th index
    # 5. remaining points follow ccw

    rect_points = np.array(rect.exterior.coords)[:-1]

    # Find major axis
    a, b, c, d = 0, 1, 2, 3
    x, y = 0, 1
    ab = gm.LineString((rect_points[a], rect_points[b]))
    bc = gm.LineString((rect_points[b], rect_points[c]))
    cd = gm.LineString((rect_points[c], rect_points[d]))
    da = gm.LineString((rect_points[d], rect_points[a]))
    edges = np.array([ab, bc, cd, da])
    longest = edges[np.argsort([edge.length for edge in edges])[-2:]]

    edge_0 = np.array(longest[0].coords)
    edge_1 = np.array(longest[1].coords)

    # Find bottom edge
    if edge_0[..., y].min() < edge_1[..., y].min():
        bottom_edge = edge_0
        top_edge = edge_1
    else:
        bottom_edge = edge_1
        top_edge = edge_0

    # arrange points starting from the left most point of the bottom
    # edge and moving counter clockwise
    if bottom_edge[0][x] > bottom_edge[1][x]:
        bottom_edge = bottom_edge[[1, 0]]
        top_edge = top_edge[[1, 0]]

    return gm.Polygon(bottom_edge.tolist() + top_edge.tolist())


def crop_rect_from_poly(
    image: Union[np.ndarray, Image.Image],
    poly: gm.Polygon,
    crop_args: Optional[Union[CropArgs, Dict]] = None,
) -> Union[np.ndarray, Image.Image]:
    if not hasattr(poly, "bounds"):
        raise ValueError("poly argument must have a .bounds attribute in order to crop")
    crop_fns = {
        False: crop_rectangle,
        True: crop_rotated_rectangle,
    }

    if crop_args is None:
        crop_args = {}

    if isinstance(crop_args, dict):
        crop_args = CropArgs(**crop_args)
    elif not isinstance(crop_args, CropArgs):
        raise ValueError(f"invalid parameter 'crop_args' expected a CropArgs object, got {crop_args}")

    kwargs = crop_args.dict()

    crop_rotated_rect = kwargs.pop(KEY_CROP_ROTATED_RECT)

    return crop_fns[crop_rotated_rect](image, poly, **kwargs)


@require_package(cv2, "cv2", "opencv")
def crop_rotated_rectangle(
    image: Union[np.ndarray, Image.Image],
    poly: gm.Polygon,
    scale: Optional[float] = None,
    pad: Optional[int] = None,
    warped_wh: Optional[Tuple[int, int]] = None,
) -> Union[np.ndarray, Image.Image]:
    """Find the minimum_rotated_rectange for the input poly and apply
    cv2.warpPerspective to transform to to rectangular crop of shape warped_wh

    see: https://shapely.readthedocs.io/en/stable/manual.html#object.minimum_rotated_rectangle

    Resulting crops will be of the same type as the input image (np.ndarray | PIL.Image)

    Args:
        image: Image to crop
        poly: Polygon to crop out
        scale: (Optional) factor applied to poly before cropping about the origin
        pad: Optional pad (in pixels) applied to poly before cropping
        warped_wh: The desired output shape of the crop
    """
    assert isinstance(pad, int) or (
        pad is None
    ), f"Croping rotated rec only support a int for 'pad', got: {pad}"

    rot_rect = poly.minimum_rotated_rectangle

    if scale is not None:
        rot_rect = affinity.scale(rot_rect, xfact=scale, yfact=scale, origin=(0, 0, 0))

    if pad is not None:
        rot_rect = rot_rect.buffer(pad, join_style=2)

    # Shapely uses a different coordinate system to us and opencv.
    # We need to make sure the points in the src_points corrispond to
    # the points in dst_points, hence this function is needed. More detail in
    # its doc string.
    rot_rect = _order_rect_points_bottom_left_ccw(rot_rect)
    src_points = np.array(list(zip(*rot_rect.exterior.coords.xy)))[:-1].astype(np.float32)

    if warped_wh is None:
        warped_wh = _get_wh_of_rotated_rect(src_points)
    w, h = warped_wh

    dst_points = np.array([[0, 0], [w, 0], [w, h], [0, h]], dtype=np.float32)

    transform_mat = cv2.getPerspectiveTransform(src_points, dst_points)

    if isinstance(image, Image.Image):
        src = np.array(image)
    elif isinstance(image, np.ndarray):
        src = image
    else:
        raise ValueError(f"Invalid image input type: {type(image)}")

    out = cv2.warpPerspective(src, transform_mat, warped_wh, flags=cv2.INTER_LINEAR)

    return out if isinstance(image, np.ndarray) else Image.fromarray(out)


def crop_rectangle(
    image: Union[np.ndarray, Image.Image],
    rect: gm.Polygon,
    scale: Optional[float] = None,
    pad: Optional[int] = None,
) -> Union[np.ndarray, Image.Image]:
    """Crop an XY Axis aligned rectangle crop image

    Resulting crops will be of the same type as the input image (np.ndarray | PIL.Image)

    Args:
        image: Image to crop
        rect: Left, Top, Right, Bottom tuple representing the rectangle to crop
        scale: Optional factor applied to rect before cropping
        pad: Optional pad (in pixels) applied to rect before cropping

    Note: Pillow.Image.crop does not give a 💩 if a crop goes out-of-bounds
    it will just add zeros as needed.
    """
    if hasattr(rect, "bounds"):
        # shapely object
        rect = rect.bounds
    else:
        raise ValueError("rect argument must have a .bounds attribute in order to crop")

    if scale is not None:
        rect = [x * scale for x in rect]

    if pad is None:
        pad = (0, 0)
    elif isinstance(pad, (float, int)):
        pad = (int(pad), int(pad))
    assert isinstance(pad, tuple)

    def pad_rect(rect, w_pad, h_pad):
        l, t, r, b = rect
        l -= w_pad
        r += w_pad
        t -= h_pad
        b += h_pad
        return (l, t, r, b)

    l, t, r, b = rect
    l, t, r, b = pad_rect((l, t, r, b), *pad)

    if isinstance(image, Image.Image):
        crop = image.crop((l, t, r, b))
    elif isinstance(image, np.ndarray):
        crop = Image.fromarray(image).crop((l, t, r, b))
        crop = np.array(crop, dtype=image.dtype)
    else:
        raise ValueError(f"Invalid image input type: {type(image)}")
    return crop



"""Networking utilities for Highlighter SDK."""

from .mdns import (
    DEFAULT_DEVICE_KEYWORDS,
    DEFAULT_SERVICE_TYPES,
    DEFAULT_TIMEOUT_SECONDS,
    DiscoveredDevice,
    DiscoveryConfig,
    DiscoveryError,
    MacResolver,
    discover_devices,
    normalize_mac,
)

__all__ = [
    "DEFAULT_DEVICE_KEYWORDS",
    "DEFAULT_SERVICE_TYPES",
    "DEFAULT_TIMEOUT_SECONDS",
    "DiscoveryConfig",
    "DiscoveredDevice",
    "DiscoveryError",
    "MacResolver",
    "normalize_mac",
    "discover_devices",
]



#!/usr/bin/env python3
#
# Usage
# ~~~~~
# LOG_LEVEL=DEBUG ./images_to_video.py

import aiko_services as aiko

FRAME_RATE = 0   # Process flat-out without delay
IMAGE_INPUT_PATHNAME = "z_input/image_{:06d}.jpg"
VIDEO_FRAME_RATE = 29.97
VIDEO_OUTPUT_PATHNAME = "z_output.mp4"

ELEMENTS_VIDEO = "aiko_services.elements.video_io"
ELEMENTS_IMAGE = "aiko_services.elements.image_io"

pipeline_definition = [
    {   "name": "ImageReadFile", "module": ELEMENTS_IMAGE,
        "successors": ["VideoWriteFile"],
        "parameters": {
            "image_pathname": IMAGE_INPUT_PATHNAME
        }
    },
    {   "name": "VideoWriteFile", "module": ELEMENTS_VIDEO,
        "parameters": {
            "video_frame_rate": VIDEO_FRAME_RATE,
            "video_pathname": VIDEO_OUTPUT_PATHNAME
        }
    }
]

# if __name__ == "__main__":
#   Pipeline_2020(pipeline_definition, FRAME_RATE).run()



# Installation (Ubuntu)
# ~~~~~~~~~~~~
# sudo apt-get install portaudio19-dev  # Provides "portaudio.h"
# pip install pyaudio sounddevice
#
# Usage
# ~~~~~
# aiko_pipeline create pipelines/audio_pipeline_0.json -s 1 -sr -ll debug
#
# aiko_pipeline create pipelines/audio_pipeline_0.json -s 1 -p rate ?????
#
# aiko_pipeline create pipelines/audio_pipeline_0.json -s 1  \
#   -p AudioReadFile.data_batch_size 8
#
# aiko_pipeline create pipelines/audio_pipeline_0.json -s 1  \
#   -p AudioReadFile.data_sources file:data_in/in_{}.mp3
#
# aiko_pipeline create pipelines/audio_pipeline_0.json -s 1  \
#   -p AudioWriteFile.path "file:data_out/out_{:02d}.mp3"
#
# aiko_pipeline create pipelines/audio_pipeline_0.json -s 1  \
#   sp AudioReadFile.data_sources file:data_in/in_00.mp3   \
#   sp AudioResample.rate ?????                              \
#   sp AudioWriteFile.data_targets file:data_out/out_00.mp3
#
# Resources
# ~~~~~~~~~
# https://people.csail.mit.edu/hubert/pyaudio/#downloads  # Install PyAudio
# https://realpython.com/playing-and-recording-sound-python
# https://stackoverflow.com/questions/62159107/python-install-correct-pyaudio
#
# To Do
# ~~~~~
# - Support .mp3, .ogg and .wav ?
# - Refactor Speech-To-Text and Text-To-Speech PipelineElements

from typing import Tuple
from pathlib import Path

import aiko_services as aiko
from aiko_services.elements.media import open_video_capture

__all__ = ["AudioOutput", "AudioReadFile"]  # "AudioWriteFile"

_LOGGER = aiko.process.logger(__name__)

_XXX_IMPORTED = False
try:
#   import xxx
    _XXX_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "audio_io.py: Couldn't import xxx module"
#   print(f"WARNING: {diagnostic}")
#   _LOGGER.warning(diagnostic)
#   raise ModuleNotFoundError(
#       'xxx package not installed.  '
#       'Install aiko_services with --extras "xxx" '
#       'or install xxx manually to use the "audio_io" module')

_CV2_IMPORTED = False
try:
    import cv2
    _CV2_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "audio_io.py: Couldn't import cv2 module"
#   print(f"WARNING: {diagnostic}")
#   _LOGGER.warning(diagnostic)
#   raise ModuleNotFoundError(
#       'opencv-python package not installed.  '
#       'Install aiko_services with --extras "opencv" '
#       'or install opencv-python manually to use the "audio_io" module')

# --------------------------------------------------------------------------- #
# Useful for Pipeline output that should be all of the audios processed

class AudioOutput(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("audio_output:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, audio_samples)  \
        -> Tuple[aiko.StreamEvent, dict]:

        return aiko.StreamEvent.OKAY, {"audio_samples": bytes}

# --------------------------------------------------------------------------- #
# AudioReadFile is a DataSource which supports ...
# - Individual audio files
# - Directory of audio files with an optional filename filter
# - TODO: Archive (tgz, zip) of audio files with an optional filename filter
#
# parameter: "data_sources" is the read file path, format variable: "audio_id"
#
# Note: Only supports Streams with "data_sources" parameter

class AudioReadFile(aiko.DataSource):  # PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("audio_read_file:0")
        context.call_init(self, "PipelineElement", context)

    def start_stream(self, stream, stream_id):
    #   stream.variables["audio_capture"] = None
        stream.variables["audio_frame_generator"] = None
        return super().start_stream(stream, stream_id, use_create_frame=False)

    def audio_frame_iterator(self, audio_capture):
        while True:
    #       status, image_bgr = audio_capture.read()
    #       if not status:
    #           break
    #       image_rgb = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2RGB)
            yield # audio_sample

    def frame_generator(self, stream, frame_id):
        audio_frame_generator = stream.variables["audio_frame_generator"]
        while True:
            if not audio_frame_generator:
                try:
                    path, file_id = next(
                        stream.variables["source_paths_generator"])
                except StopIteration:
                    stream.variables["source_paths_generator"] = None
                    diagnostic = "End of audio file(s)"
                    return aiko.StreamEvent.STOP, {"diagnostic": diagnostic}

                if not path.is_file():
                    diagnostic = f'path "{path}" must be a file'
                    return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

            # TODO: handle integer audio "path" ... maybe ?
            #   if isinstance(path, str) and path.isdigit():
            #       path = int(str(path))

                audio_capture = open_video_capture(str(path))
                if not audio_capture.isOpened():
                    diagnostic = f"Couldn't open audio file: {path}"
                    return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}
                stream.variables["audio_capture"] = audio_capture
                audio_frame_generator = self.audio_frame_iterator(audio_capture)
                stream.variables["audio_frame_generator"] =  \
                    audio_frame_generator

            try:
                image_rgb = next(audio_frame_generator)
                return aiko.StreamEvent.OKAY, {"images": [image_rgb]}
            except StopIteration:
                audio_frame_generator = None
                stream.variables["audio_frame_generator"] = None

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        self.logger.debug(f"{self.my_id()}")
        return aiko.StreamEvent.OKAY, {"images": images}

    def stop_stream(self, stream, stream_id):
        audio_capture = stream.variables["audio_capture"]
        if audio_capture.isOpened():
            audio_capture.release()
            stream.variables["audio_capture"] = None
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #
"""
# Usage
# ~~~~~
# aiko_pipeline create ../../examples/pipeline/pipeline_mic_fft_graph.json
#
# On "Spectrum" window, press "x" to exit
#
# To Do
# ~~~~~
# - Ensure that Registar interaction occurs prior to flooding the Pipeline !
# - Implement Microphone PipelineElement start_stream(), move __init__() code

from hashlib import md5
from io import BytesIO
import numpy as np
from typing import Tuple
import zlib

import aiko_services as aiko
from aiko_services.main.utilities import get_namespace, parse_number

__all__ = [
    "PE_AudioFilter", "PE_AudioResampler",
    "PE_FFT", "PE_GraphXY",
    "PE_MicrophonePA", "PE_MicrophoneSD", "PE_Speaker"
]

_LOGGER = aiko.logger(__name__)

# PipelineElement parameters
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# PE_MicrophonePA, PE_MicrophoneSD
AUDIO_CHANNELS = 1  # 1 or 2 channels

# --------------------------------------------------------------------------- #
# TODO: Turn some of these literals into PipelineElement parameters
# TODO: Place some of these PipelineElement parameters into self.share[]
# TODO: Update via ECProducer

AF_AMPLITUDE_MINIMUM = 0.1
AF_AMPLITUDE_MAXIMUM = 12
AF_FREQUENCY_MINIMUM = 10    # Hertz
AF_FREQUENCY_MAXIMUM = 9000  # Hertz
AF_SAMPLES_MAXIMUM = 100

class PE_AudioFilter(PipelineElement):
    def __init__(self, context):
        context.set_protocol("audio_filter:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self,
        stream, amplitudes, frequencies) -> Tuple[aiko.StreamEvent, dict]:

        data = list(zip(np.abs(frequencies), amplitudes))
        data = [(x,y) for x,y in data if y >= AF_AMPLITUDE_MINIMUM]
        data = [(x,y) for x,y in data if x >= AF_FREQUENCY_MINIMUM]
        data = [(x,y) for x,y in data if y <= AF_AMPLITUDE_MAXIMUM]
        data = [(x,y) for x,y in data if x <= AF_FREQUENCY_MAXIMUM]
        amplitude_key = 1
        data = sorted(data, key=lambda x: x[amplitude_key], reverse=True)
        data = data[:AF_SAMPLES_MAXIMUM]
        if len(data):
            frequencies, amplitudes = zip(*data)
        else:
            frequencies, amplitudes = [], []

        return aiko.StreamEvent.AIKO,
               {"amplitudes": amplitudes, "frequencies": frequencies}

# --------------------------------------------------------------------------- #
# TODO: Turn some of these literals into PipelineElement parameters
# TODO: Place some of these PipelineElement parameters into self.share[]
# TODO: Update via ECProducer

AR_BAND_COUNT = 8

class PE_AudioResampler(PipelineElement):  # TODO: AudioTransform ?
    def __init__(self, context):
        context.set_protocol("audio_resample:0")
        context.call_init(self, "PipelineElement", context)
        self.counter = 0

# Frequencies: [0:2399] --> 0, 10, 20, ... 23990 Hz
# Extract 0 to 8000 Hz
# Consolidate that into 8 bands !
# Check the maths and the normalization !

    def process_frame(self,
        stream, amplitudes, frequencies) -> Tuple[aiko.StreamEvent, dict]:

        amplitudes = amplitudes[0:len(amplitudes) // 2]     # len: 2400
        frequencies = frequencies[0:len(frequencies) // 2]  # len: 2400

        frequency_range = frequencies[-1] - frequencies[0]  # 23990.0
        band_width = frequency_range / AR_BAND_COUNT / 10      # 299.875  # TODO: MAGIC NUMBER !!
        band_frequencies = []
        band_amplitudes = []

        for band in range(AR_BAND_COUNT):
            band_start = band * band_width
            band_end = band_start + band_width

        # TODO: Easier to just use indices / slices, rather than a mask ?
            mask = (frequencies >= band_start) & (frequencies < band_end)
            amplitudes_sum = np.sum(amplitudes[mask])

            band_frequency_count = np.sum(mask)
            normalized_amplitudes_sum = amplitudes_sum / band_frequency_count

            band_frequencies.append((band_start + band_end) / 2)
            band_amplitudes.append(amplitudes_sum)
        #   band_amplitudes.append(normalized_amplitudes_sum)

        frequencies = np.array(band_frequencies)
        amplitudes = np.array(band_amplitudes)

        topic_path = "aiko/esp32_ed6cxc/0/0/in"
        self.counter += 1
        if self.counter % 5:
            aiko.message.publish(topic_path, "(led:fill 0 0 0)")
            x = 0
            for frequency, amplitude in zip(frequencies, amplitudes):
                print(f"Band: {frequency:.0f} Hz, amplitude: {amplitude:.4f}")
                a = f"{amplitude:.0f}"
                payload_out = f"(led:line 255 0 0 {x} 0 {x} {a})"
                aiko.message.publish(topic_path, payload_out)
                x += 1
            aiko.message.publish(topic_path, "(led:write)")

        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #
# TODO: Turn some of these literals into PipelineElement parameters
# TODO: Place some of these PipelineElement parameters into self.share[]
# TODO: Update via ECProducer

FFT_AMPLITUDE_SCALER = 1_000_000

class PE_FFT(PipelineElement):
    def __init__(self, context):
        context.set_protocol("fft:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream, audio) -> Tuple[aiko.StreamEvent, dict]:
        fft_output = np.fft.fft(audio)

        amplitudes = np.abs(fft_output / FFT_AMPLITUDE_SCALER)
        top_index = np.argmax(amplitudes)

        frequencies = np.fft.fftfreq(
            PA_AUDIO_CHUNK_SIZE, 1 / PA_AUDIO_SAMPLE_RATE)
        top_amplitude = int(amplitudes[top_index])
        top_frequency = np.abs(frequencies[top_index])
        _LOGGER.debug(
            f"{self.my_id()} Loudest: {top_frequency} Hz: {top_amplitude}")

        return aiko.StreamEvent.OKAY,
               {"amplitudes": amplitudes, "frequencies": frequencies}

# --------------------------------------------------------------------------- #
# TODO: Turn some of these literals into PipelineElement parameters
# TODO: Place some of these PipelineElement parameters into self.share[]
# TODO: Update via ECProducer

import cv2
import io
from PIL import Image
import pygal

GRAPH_FRAME_PERIOD = 1  # milliseconds
GRAPH_TITLE = "Spectrum"
WINDOW_TITLE = GRAPH_TITLE

class PE_GraphXY(PipelineElement):
    def __init__(self, context):
        context.set_protocol("graph_xy:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self,
        stream, amplitudes, frequencies) -> Tuple[aiko.StreamEvent, dict]:

        data = list(zip(np.abs(frequencies), amplitudes))

        graph = pygal.XY(
            x_title="Frequency (Hz)", y_title="Amplitude", stroke=False)
        graph.title = GRAPH_TITLE
    #   graph.x_limit = [0, AF_FREQUENCY_MAXIMUM]  # Fails :(
    #   graph.y_limit = [0, AF_AMPLITUDE_MAXIMUM]  # Fails :(
        graph.add("Audio", data)
        graph.add("Limit",
            [(0, 0), (AF_FREQUENCY_MAXIMUM, AF_AMPLITUDE_MAXIMUM)])

        memory_file = io.BytesIO()
        graph.render_to_png(memory_file)
        memory_file.seek(0)
        image = Image.open(memory_file)
        image_bgr = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
        cv2.imshow(WINDOW_TITLE, image_bgr)
        if cv2.waitKey(GRAPH_FRAME_PERIOD) & 0xff == ord('x'):
            return False, {}
        else:
            return aiko.StreamEvent.OKAY,
                   {"amplitudes": amplitudes, "frequencies": frequencies}

# --------------------------------------------------------------------------- #
# TODO: Turn some of these literals into PipelineElement parameters
# TODO: Place some of these PipelineElement parameters into self.share[]
# TODO: Update via ECProducer

import pyaudio
from threading import Thread

PA_AUDIO_FORMAT = pyaudio.paInt16

PA_AUDIO_SAMPLE_RATE = 16000       # voice 16,000 or 44,100 or 48,000 Hz
# PA_AUDIO_SAMPLE_RATE = 48000     # music / spectrum analyser

PA_AUDIO_CHUNK_SIZE = PA_AUDIO_SAMPLE_RATE * 2          # Voice: 2.0 seconds
# PA_AUDIO_CHUNK_SIZE = int(PA_AUDIO_SAMPLE_RATE / 10)  # FFT:   0.1 seconds

def py_error_handler(filename, line, function, err, fmt):
    pass

from ctypes import *

def hide_alsa_messages():
    ERROR_HANDLER_FUNC = CFUNCTYPE(
        None, c_char_p, c_int, c_char_p, c_int, c_char_p)
    c_error_handler = ERROR_HANDLER_FUNC(py_error_handler)
    asound = cdll.LoadLibrary("libasound.so")
    asound.snd_lib_error_set_handler(c_error_handler)

import os
import platform
import sys

def pyaudio_initialize():
    if platform.system() == "Linux":
        hide_alsa_messages()  # Fails on Mac OS X and Windows
    devnull = os.open(os.devnull, os.O_WRONLY)
    old_stderr = os.dup(2)
    sys.stderr.flush()
    os.dup2(devnull, 2)
    os.close(devnull)

    py_audio = pyaudio.PyAudio()

    os.dup2(old_stderr, 2)
    os.close(old_stderr)
    return py_audio

class PE_MicrophonePA(PipelineElement):
    def __init__(self, context):
        context.set_protocol("microphone:0")
        context.call_init(self, "PipelineElement", context)

        self.share["frame_id"] = -1

        audio_channels, _ = self.get_parameter("audio_channels", AUDIO_CHANNELS)
        self.py_audio = pyaudio_initialize()
        self.audio_stream = self.py_audio.open(
            channels=audio_channels,
            format=PA_AUDIO_FORMAT,
            frames_per_buffer=PA_AUDIO_CHUNK_SIZE,
            input=True,
            rate=PA_AUDIO_SAMPLE_RATE)
        self.pipeline.create_stream(0)
        self.thread = Thread(target=self._audio_run, daemon=True).start()

    def _audio_run(self):
        self.terminate = False
        while not self.terminate:
            audio_sample_raw = self.audio_stream.read(PA_AUDIO_CHUNK_SIZE)
            audio_sample = np.frombuffer(audio_sample_raw, dtype=np.int16)

            frame_id = self.share["frame_id"] + 1
            self.ec_producer.update("frame_id", frame_id)
            stream = {"stream_id": 0, "frame_id": frame_id}
            self.create_frame(stream, {"audio": audio_sample})

        self.audio_stream.close()
        self.py_audio.terminate()

    def process_frame(self, stream, audio) -> Tuple[aiko.StreamEvent, dict]:
    #   _LOGGER.debug(f"{self.my_id()} len(audio): {len(audio)}")
        return aiko.StreamEvent.OKAY, {"audio": audio}

    def stop_stream(self, stream, stream_id):
        _LOGGER.debug(f"{self.my_id()}: stop_stream()")
        self.terminate = True
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #

SD_AUDIO_CHUNK_DURATION = 5.0    # voice: audio chunk duration in seconds
# SD_AUDIO_CHUNK_DURATION = 0.1  # music / spectrum analyser

SD_AUDIO_SAMPLE_DURATION = 5.0   # audio sample size to process
SD_AUDIO_SAMPLE_RATE = 16000     # voice 16,000 or 44,100 or 48,000 Hz
# SD_AUDIO_SAMPLE_RATE = 48000   # music / spectrum analyser

SD_SAMPLES_PER_CHUNK = SD_AUDIO_SAMPLE_RATE * SD_AUDIO_CHUNK_DURATION

import sounddevice as sd

class PE_MicrophoneSD(PipelineElement):
    def __init__(self, context):
        context.set_protocol("microphone:0")
        context.call_init(self, "PipelineElement", context)

        self.share["mute"] = 0
        self.share["frame_id"] = -1
        self.pipeline.create_stream(0)
        self._time_mute = 0
        self._thread = Thread(target=self._audio_run, daemon=True).start()

    def _audio_run(self):
        self.terminate = False
        self._audio_sampler_start()
        audio_channels, _ = self.get_parameter("audio_channels", AUDIO_CHANNELS)

        with sd.InputStream(callback=self._audio_sampler,
            channels=audio_channels, samplerate=SD_AUDIO_SAMPLE_RATE):

            while not self.terminate:
                sd.sleep(int(SD_AUDIO_CHUNK_DURATION * 1000))

    def _audio_sampler_start(self):
        frame_id = self.share["frame_id"] + 1
        self.ec_producer.update("frame_id", frame_id)
        self._audio_sample = np.empty((0, 1), dtype=np.float32)
        return frame_id - 1

    def _audio_sampler(self, indata, frames, time_, status):
        if status:
            _LOGGER.error(f"SoundDevice error: {status}")
        else:
        #   _LOGGER.debug(f"SoundDevice callback: {len(indata)} bytes")
            if self._time_mute == 0 or time.monotonic() > self._time_mute:
                if self._time_mute:
                    self._time_mute = 0
                    self.ec_producer.update("mute", 0)

                indata = indata.copy().astype(np.float32)
                self._audio_sample = np.concatenate(
                    (self._audio_sample, indata), axis=0)
                if len(self._audio_sample) > SD_SAMPLES_PER_CHUNK:
                    audio_sample = self._audio_sample
                    frame_id = self._audio_sampler_start()
                    stream = {"stream_id": 0, "frame_id": frame_id}
                    self.create_frame(stream, {"audio": audio_sample})

    def mute(self, duration):
        duration = parse_number(duration)
        time_mute = time.monotonic() + duration
        if duration == 0 or time_mute >= self._time_mute:
            self._audio_sample = np.empty((0, 1), dtype=np.float32)
            self._time_mute = time_mute
            self.ec_producer.update("mute", duration)

    def process_frame(self, stream, audio) -> Tuple[aiko.StreamEvent, dict]:
        audio_channels, _ = self.get_parameter("audio_channels", AUDIO_CHANNELS)
        _LOGGER.info(f"SoundDevice audio_channels: {audio_channels}")
        return aiko.StreamEvent.OKAY, {"audio": audio}

    def stop_stream(self, stream, stream_id):
        _LOGGER.debug(f"{self.my_id()}: stop_stream()")
        self.terminate = True
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #

import time

# SP_AUDIO_SAMPLE_RATE = PA_AUDIO_SAMPLE_RATE
# SP_AUDIO_SAMPLE_RATE = SD_AUDIO_SAMPLE_RATE
SP_AUDIO_SAMPLE_RATE = 22050                   # coqui.ai text-to-speech
SP_SPEED_UP = 0.7                              # talk time fine tuning !

class PE_Speaker(PipelineElement):
    def __init__(self, context):
        context.set_protocol("speaker:0")
        context.call_init(self, "PipelineElement", context)

        self._microphone_service = None
        do_discovery(
            PE_MicrophoneSD,
            ServiceFilter("*", "*", "microphone:0", "*", "*", "*"),
                lambda _, service: self._discovery_handler(service),
                lambda _:          self._discovery_handler(None))

    def _discovery_handler(self, service):
        self._microphone_service = service

    def process_frame(self, stream, audio) -> Tuple[aiko.StreamEvent, dict]:
        if audio is not None:
            sd.play(audio, SP_AUDIO_SAMPLE_RATE)
            duration = len(audio) / SD_AUDIO_SAMPLE_RATE * SP_SPEED_UP
            if self._microphone_topic_path:
                self._microphone_topic_path.mute(duration)
            time.sleep(duration)
        return aiko.StreamEvent.OKAY, {"audio": audio}
"""
# --------------------------------------------------------------------------- #
